"""Contains functions pertaining to the design of physical and chemical unit
processes of AguaClara water treatment plants.
"""

from aguaclara.core.units import u
import aguaclara.core.constants as con
import aguaclara.core.utility as ut
import aguaclara.core.pipes as pipe

import numpy as np
from scipy import interpolate, integrate
import warnings

############################ Gas ##############################


@ut.list_handler()
def density_air(Pressure, MolarMass, Temperature):
    """
    .. deprecated::
        `density_air` is deprecated; use `density_gas` instead.
    """
    warnings.warn('density_air is deprecated; use density_gas instead.',
                  UserWarning)
    return density_gas(Pressure, MolarMass, Temperature)


@ut.list_handler()
def density_gas(Pressure, MolarMass, Temperature):
    """Return the density of air at the given pressure, molar mass, and
    temperature.

    :param Pressure: pressure of air in the system
    :type Pressure: u.pascal
    :param MolarMass: molar mass of air in the system
    :type MolarMass: u.gram/u.mol
    :param Temperature: Temperature of air in the system
    :type Temperature: u.degK

    :return: density of air in the system
    :rtype: u.kg/u.m**3
    """
    return (Pressure * MolarMass / (u.R * Temperature)).to(u.kg/u.m**3)

########################## Geometry ###########################


@ut.list_handler()
def area_circle(DiamCircle):
    """Return the area of a circle given its diameter.

    :param DiamCircle: diameter of circle
    :type DiamCircle: u.m

    :return: area of circle
    :rtype: u.m**2
    """
    ut.check_range([DiamCircle.magnitude, ">0", "DiamCircle"])
    return (np.pi / 4 * DiamCircle**2)


@ut.list_handler()
def diam_circle(AreaCircle):
    """Return the diameter of a circle given its area.

    :param AreaCircle: area of circle
    :type AreaCircle: u.m**2

    :return: diameter of circle
    :rtype: u.m
    """

    ut.check_range([AreaCircle.magnitude, ">0", "AreaCircle"])
    return np.sqrt(4 * AreaCircle / np.pi)

####################### Water Properties #######################


#:
RE_TRANSITION_PIPE = 2100

#: Table of temperatures and the corresponding water density.
#:
#: WATER_DENSITY_TABLE[0] is a list of water temperatures, in Kelvin.
#: WATER_DENSITY_TABLE[1] is the corresponding densities, in kg/m³.
WATER_DENSITY_TABLE = [(273.15, 278.15, 283.15, 293.15, 303.15, 313.15,
                        323.15, 333.15, 343.15, 353.15, 363.15, 373.15
                        ), (999.9, 1000, 999.7, 998.2, 995.7, 992.2,
                            988.1, 983.2, 977.8, 971.8, 965.3, 958.4
                            )
                       ]


@ut.list_handler()
def viscosity_dynamic(temp):
    """
    .. deprecated::
        `viscosity_dynamic` is deprecated; use `viscosity_dynamic_water`
        instead.
    """
    warnings.warn('viscosity_dynamic is deprecated; use '
                  'viscosity_dynamic_water instead.', UserWarning)
    return viscosity_dynamic_water(temp)


@ut.list_handler()
def viscosity_dynamic_water(Temperature):
    """Return the dynamic viscosity of water at a given temperature.

    :param Temperature: temperature of water
    :type Temperature: u.degK

    :return: dynamic viscosity of water
    :rtype: u.kg/(u.m*u.s)
    """
    ut.check_range([Temperature.magnitude, ">=0", "Temperature in Kelvin"])
    return 2.414 * (10**-5) * u.kg/(u.m*u.s) * 10**(247.8*u.degK /
                                                    (Temperature - 140*u.degK))


@ut.list_handler()
def density_water(Temperature=None, *, temp=None):
    """Return the density of water at a given temperature.

    :param Temperature: temperature of water
    :type Temperature: u.degK

    :param temp: deprecated; use Temperature instead

    :return: density of water
    :rtype: u.kg/u.m**3
    """
    if Temperature is not None and temp is not None:
        raise TypeError("density_water received both Temperature and temp")
    elif Temperature is None and temp is None:
        raise TypeError("density_water missing Temperature argument")
    elif temp is not None:
        warnings.warn("temp is deprecated; use Temperature instead.",
                      UserWarning)
        Temperature = temp

    ut.check_range([Temperature.magnitude, ">=0", "Temperature in Kelvin"])
    rhointerpolated = interpolate.CubicSpline(WATER_DENSITY_TABLE[0],
                                              WATER_DENSITY_TABLE[1])
    Temperature = Temperature.to(u.degK).magnitude
    return rhointerpolated(Temperature).item() * u.kg/u.m**3


@ut.list_handler()
def viscosity_kinematic(temp):
    """
    .. deprecated::
        `viscosity_kinematic` is deprecated; use `viscosity_kinematic_water`
        instead.
    """
    warnings.warn('viscosity_kinematic is deprecated; use '
                  'viscosity_kinematic_water instead.', UserWarning)
    return viscosity_kinematic_water(temp)


@ut.list_handler()
def viscosity_kinematic_water(Temperature):
    """Return the kinematic viscosity of water at a given temperature.

    :param Temperature: temperature of water
    :type Temperature: u.degK

    :return: kinematic viscosity of water
    :rtype: u.m**2/u.s
    """
    ut.check_range([Temperature.magnitude, ">=0", "Temperature in Kelvin"])
    return (viscosity_dynamic_water(Temperature) / density_water(Temperature))

####################### Hydraulic Radius #######################


@ut.list_handler()
def radius_hydraulic(Width, Depth, openchannel):
    """
    .. deprecated::
        `radius_hydraulic` is deprecated; use `radius_hydraulic_rect` instead.
    """
    warnings.warn('radius_hydraulic is deprecated; use radius_hydraulic_rect '
                  'instead.', UserWarning)
    return radius_hydraulic_rect(Width, Depth, openchannel)


@ut.list_handler()
def radius_hydraulic_rect(Width, Depth, OpenChannel):
    """Return the hydraulic radius of a rectangular channel given width and
    depth of water.

    :param Width: width of channel
    :type Width: u.m
    :param Depth: depth of water in channel
    :type Depth: u.m
    :param OpenChannel: true if channel is open, false if closed
    :type OpenChannel: boolean

    :return: hydraulic radius of rectangular channel
    :rtype: u.m
    """
    ut.check_range([Width.magnitude, ">0", "Width"],
                   [Depth.magnitude, ">0", "Depth"],
                   [OpenChannel, "boolean", "OpenChannel"])
    if OpenChannel:
        return ((Width*Depth) / (Width + 2*Depth))
    else:
        return ((Width*Depth) / (2 * (Width+Depth)))


@ut.list_handler()
def radius_hydraulic_general(Area, PerimWetted):
    """
    .. deprecated::
        `radius_hydraulic_general` is deprecated; use
        `radius_hydraulic_channel` instead.
    """
    warnings.warn('radius_hydraulic_general is deprecated; use '
                  'radius_hydraulic_channel instead.', UserWarning)
    return radius_hydraulic_channel(Area, PerimWetted)


@ut.list_handler()
def radius_hydraulic_channel(Area, PerimWetted):
    """Return the hydraulic radius of a general channel given cross sectional
    area and wetted perimeter.

    :param Area: cross sectional area of channel
    :type Area: u.m**2
    :param PerimWetted: wetted perimeter of channel
    :type PerimWetted: u.m

    :return: hydraulic radius of general channel
    :rtype: u.m
    """
    ut.check_range([Area.magnitude, ">0", "Area"],
                   [PerimWetted.magnitude, ">0", "Wetted perimeter"])
    return (Area / PerimWetted)

####################### Reynolds Number #######################


@ut.list_handler()
def re_pipe(FlowRate, Diam, Nu):
    """Return the Reynolds number of flow through a pipe.

    :param FlowRate: flow rate through pipe
    :type FlowRate: u.m**3/u.s
    :param Diam: diameter of pipe
    :type Diam: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s

    :return: Reynolds number of flow through pipe
    :rtype: u.dimensionless
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Diam.magnitude, ">0", "Diameter"],
                   [Nu.magnitude, ">0", "Nu"])
    return ((4 * FlowRate) / (np.pi * Diam * Nu)).to(u.dimensionless)


@ut.list_handler()
def re_rect(FlowRate, Width, Depth, Nu, OpenChannel=None, *, openchannel=None):
    """Return the Reynolds number of flow through a rectangular channel.

    :param FlowRate: flow rate through channel
    :type FlowRate: u.m**3/u.s
    :param Width: width of channel
    :type Width: u.m
    :param Depth: depth of water in channel
    :type Depth: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param OpenChannel: true if channel is open, false if closed
    :type OpenChannel: boolean

    :param openchannel: deprecated; use OpenChannel instead

    :return: Reynolds number of flow through rectangular channel
    :rtype: u.dimensionless
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Nu.magnitude, ">0", "Nu"])
    if OpenChannel is not None and openchannel is not None:
        raise TypeError("re_rect received both OpenChannel and openchannel")
    elif OpenChannel is None and openchannel is None:
        raise TypeError("re_rect missing OpenChannel argument")
    elif openchannel is not None:
        warnings.warn("openchannel is deprecated; use OpenChannel instead.",
                      UserWarning)
        OpenChannel = openchannel

    return (4 * FlowRate * radius_hydraulic_rect(Width, Depth, OpenChannel)
            / (Width * Depth * Nu)).to(u.dimensionless)


@ut.list_handler()
def re_general(Vel, Area, PerimWetted, Nu):
    """
    .. deprecated::
        `re_general` is deprecated; use `re_channel` instead.
    """
    warnings.warn('re_general is deprecated; use re_channel instead.',
                  UserWarning)
    return re_channel(Vel, Area, PerimWetted, Nu)


@ut.list_handler()
def re_channel(Vel, Area, PerimWetted, Nu):
    """Return the Reynolds number of flow through a general cross section.

    :param Vel: velocity of fluid
    :type Vel: u.m/u.s
    :param Area: cross sectional area of channel
    :type Area: u.m**2
    :param PerimWetted: wetted perimeter of channel
    :type PerimWetted: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s

    :return: Reynolds number of flow through general cross section
    :rtype: u.dimensionless
    """
    ut.check_range([Vel.magnitude, ">=0", "Velocity"],
                   [Nu.magnitude, ">0", "Nu"])
    return (4 * radius_hydraulic_channel(Area, PerimWetted) * Vel / Nu).to(u.dimensionless)

########################### Friction ###########################


@ut.list_handler()
def fric(FlowRate, Diam, Nu, PipeRough):
    """
    .. deprecated::
        `fric` is deprecated; use `fric_pipe` instead.
    """
    warnings.warn('fric is deprecated; use fric_pipe instead', UserWarning)
    return fric_pipe(FlowRate, Diam, Nu, PipeRough)


@ut.list_handler()
def fric_pipe(FlowRate, Diam, Nu, Roughness):
    """Return the friction factor for pipe flow.

    For laminar flow, the friction factor is 64 is divided the Reynolds number.
    For turbulent flows, friction factor is calculated using the Swamee-Jain
    equation, which works best for Re > 3000 and ε/Diam < 0.02.

    :param FlowRate: flow rate through pipe
    :type FlowRate: u.m**3/u.s
    :param Diam: diameter of pipe
    :type Diam: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of pipe
    :type Roughness: u.m

    :return: friction factor of flow through pipe
    :rtype: u.dimensionless
    """
    ut.check_range([Roughness.magnitude, ">=0", "Pipe roughness"])
    if re_pipe(FlowRate, Diam, Nu) >= RE_TRANSITION_PIPE:
        f = (0.25 / (np.log10(Roughness / (3.7 * Diam)
                              + 5.74 / re_pipe(FlowRate, Diam, Nu) ** 0.9
                              )
                     ) ** 2
             )
    else:
        f = 64 / re_pipe(FlowRate, Diam, Nu)
    return f * u.dimensionless


@ut.list_handler()
def fric_rect(FlowRate, Width, Depth, Nu, Roughness=None, OpenChannel=None, *,
              PipeRough=None, openchannel=None):
    """Return the friction factor of a rectangular channel.

    The Swamee-Jain equation is adapted for a rectangular channel.

    :param FlowRate: flow rate through channel
    :type FlowRate: u.m**3/u.s
    :param Width: width of channel
    :type Width: u.m
    :param Depth: depth of water in channel
    :type Depth: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of channel
    :type Roughness: u.m
    :param OpenChannel: true if channel is open, false if closed
    :type OpenChannel: boolean

    :param PipeRough: deprecated; use Roughness instead
    :param openchannel: deprecated; use OpenChannel instead

    :return: friction factor of flow through rectangular channel
    :rtype: u.dimensionless
    """
    if Roughness is not None and PipeRough is not None:
        raise TypeError("fric_rect received both Roughness and PipeRough")
    elif Roughness is None and PipeRough is None:
        raise TypeError("fric_rect missing Roughness argument")
    elif OpenChannel is not None and openchannel is not None:
        raise TypeError("fric_rect received both OpenChannel and openchannel")
    elif OpenChannel is None and openchannel is None:
        raise TypeError("fric_rect missing OpenChannel argument")
    else:
        if PipeRough is not None:
            warnings.warn("PipeRough is deprecated; use Roughness instead.",
                          UserWarning)
            Roughness = PipeRough
        if openchannel is not None:
            warnings.warn("openchannel is deprecated; use OpenChannel instead.",
                          UserWarning)
            OpenChannel = openchannel

    ut.check_range([Roughness.magnitude, ">=0", "Pipe roughness"])
    if re_rect(FlowRate, Width, Depth, Nu, OpenChannel) >= RE_TRANSITION_PIPE:
        # Diam = 4*R_h in adapted Swamee-Jain equation
        return (0.25 * u.dimensionless
                / (np.log10((Roughness
                             / (3.7 * 4
                                * radius_hydraulic_rect(Width, Depth,
                                                        OpenChannel)
                                )
                             )
                            + (5.74 / (re_rect(FlowRate, Width, Depth,
                                               Nu, OpenChannel) ** 0.9)
                               )
                            )
                   ) ** 2
                )
    else:
        return 64 * u.dimensionless / re_rect(FlowRate, Width, Depth, Nu,
                                              OpenChannel)


@ut.list_handler()
def fric_general(Area, PerimWetted, Vel, Nu, PipeRough):
    """
    .. deprecated::
        `fric_general` is deprecated; use `fric_channel` instead.
    """
    warnings.warn('fric_general is deprecated; use fric_channel instead.',
                  UserWarning)
    return fric_channel(Area, PerimWetted, Vel, Nu, PipeRough)


@ut.list_handler()
def fric_channel(Area, PerimWetted, Vel, Nu, Roughness):
    """Return the friction factor for a general channel.

    The Swamee-Jain equation is adapted for a general cross-section.

    :param Area: cross sectional area of channel
    :type Area: u.m**2
    :param PerimWetted: wetted perimeter of channel
    :type PerimWetted: u.m
    :param Vel: velocity of fluid
    :type Vel: u.m/u.s
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param PipeRough: roughness of channel
    :type PipeRough: u.m

    :return: friction factor for flow through general channel
    :rtype: u.dimensionless
    """
    ut.check_range([Roughness.magnitude, ">=0", "Pipe roughness"])
    if re_channel(Vel, Area, PerimWetted, Nu) >= RE_TRANSITION_PIPE:
        # Diam = 4*R_h in adapted Swamee-Jain equation
        f = (0.25 /
             (np.log10((Roughness
                        / (3.7 * 4
                           * radius_hydraulic_channel(Area, PerimWetted)
                           )
                        )
                       + (5.74
                          / re_channel(Vel, Area, PerimWetted, Nu) ** 0.9
                          )
                       )
              ) ** 2
             )
    else:
        f = 64 / re_channel(Vel, Area, PerimWetted, Nu)
    return f * u.dimensionless

######################### Head Loss #########################


@ut.list_handler()
def headloss_fric(FlowRate, Diam, Length, Nu, PipeRough):
    """
    .. deprecated::
        `headloss_fric` is deprecated; use `headloss_major_pipe` instead.
    """
    warnings.warn('headloss_fric is deprecated; use headloss_major_pipe instead',
                  UserWarning)
    return headloss_major_pipe(FlowRate, Diam, Length, Nu, PipeRough)


@ut.list_handler()
def headloss_major_pipe(FlowRate, Diam, Length, Nu, Roughness):
    """Return the major head loss (due to wall shear) in a pipe.

    This function applies to both laminar and turbulent flows.

    :param FlowRate: flow rate through pipe
    :type FlowRate: u.m**3/u.s
    :param Diam: diameter of pipe
    :type Diam: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of pipe
    :type Roughness: u.m

    :return: major head loss in pipe
    :rtype: u.m
    """
    ut.check_range([Length.magnitude, ">0", "Length"])
    return (fric_pipe(FlowRate, Diam, Nu, Roughness)
            * 8 / (u.gravity * np.pi**2)
            * (Length * FlowRate**2) / Diam**5
            ).to(u.m)


@ut.list_handler()
def headloss_exp(FlowRate, Diam, KMinor):
    """
    .. deprecated::
        `headloss_exp` is deprecated; use `headloss_minor_pipe` instead.
    """
    warnings.warn('headloss_exp is deprecated; use headloss_minor_pipe instead',
                  UserWarning)
    return headloss_minor_pipe(FlowRate, Diam, KMinor)


@ut.list_handler()
def headloss_minor_pipe(FlowRate, Diam, KMinor):
    """Return the minor head loss (due to changes in geometry) in a pipe.

    This function applies to both laminar and turbulent flows.

    :param FlowRate: flow rate through pipe
    :type FlowRate: u.m**3/u.s
    :param Diam: diameter of pipe
    :type Diam: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: minor head loss in pipe
    :rtype: u.m
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Diam.magnitude, ">0", "Diameter"],
                   [KMinor, ">=0", "K minor"])
    return (KMinor * 8 / (u.gravity * np.pi**2) * FlowRate**2 / Diam**4).to(u.m)


@ut.list_handler()
def headloss(FlowRate, Diam, Length, Nu, PipeRough, KMinor):
    """
    .. deprecated::
        `headloss` is deprecated; use `headloss_pipe` instead.
    """
    warnings.warn('headloss is deprecated; use headloss_pipe instead',
                  UserWarning)
    return headloss_pipe(FlowRate, Diam, Length, Nu, PipeRough, KMinor)


@ut.list_handler()
def headloss_pipe(FlowRate, Diam, Length, Nu, Roughness, KMinor):
    """Return the total head loss from major and minor losses in a pipe.

    This function applies to both laminar and turbulent flows.

    :param FlowRate: flow rate through pipe
    :type FlowRate: u.m**3/u.s
    :param Diam: diameter of pipe
    :type Diam: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of pipe
    :type Roughness: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: total head loss in pipe
    :rtype: u.m
    """
    return (headloss_major_pipe(FlowRate, Diam, Length, Nu, Roughness)
            + headloss_minor_pipe(FlowRate, Diam, KMinor))


@ut.list_handler()
def headloss_fric_rect(FlowRate, Width, Depth, Length, Nu, PipeRough, openchannel):
    """
    .. deprecated::
        `headloss_fric_rect` is deprecated; use `headloss_major_rect` instead.
    """
    warnings.warn('headloss_fric_rect is deprecated; use headloss_major_rect instead',
                  UserWarning)
    return headloss_major_rect(FlowRate, Width, Depth, Length, Nu, PipeRough, openchannel)


@ut.list_handler()
def headloss_major_rect(FlowRate, Width, Depth, Length, Nu, Roughness, OpenChannel):
    """Return the major head loss due to wall shear in a rectangular channel.

    This equation applies to both laminar and turbulent flows.

    :param FlowRate: flow rate through channel
    :type FlowRate: u.m**3/u.s
    :param Width: width of channel
    :type Width: u.m
    :param Depth: depth of water in channel
    :type Depth: u.m
    :param Length: length of channel
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of channel
    :type Roughness: u.m
    :param OpenChannel: true if channel is open, false if closed
    :type OpenChannel: boolean

    :return: major head loss in rectangular channel
    :rtype: u.m
    """
    ut.check_range([Length.magnitude, ">0", "Length"])
    return (fric_rect(FlowRate, Width, Depth, Nu,
                      Roughness, OpenChannel)
            * Length
            / (4 * radius_hydraulic_rect(Width, Depth, OpenChannel))
            * FlowRate**2
            / (2 * u.gravity * (Width*Depth)**2)
            ).to(u.m)


@ut.list_handler()
def headloss_exp_rect(FlowRate, Width, Depth, KMinor):
    """
    .. deprecated::
        `headloss_exp_rect` is deprecated; use `headloss_minor_rect` instead.
    """
    warnings.warn('headloss_exp_rect is deprecated; use headloss_minor_rect instead',
                  UserWarning)
    return headloss_minor_rect(FlowRate, Width, Depth, KMinor)


@ut.list_handler()
def headloss_minor_rect(FlowRate, Width, Depth, KMinor):
    """Return the minor head loss due to expansion in a rectangular channel.

    This equation applies to both laminar and turbulent flows.

    :param FlowRate: flow rate through channel
    :type FlowRate: u.m**3/u.s
    :param Width: width of channel
    :type Width: u.m
    :param Depth: depth of water in channel
    :type Depth: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: minor head loss in rectangular channel
    :rtype: u.m
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Width.magnitude, ">0", "Width"],
                   [Depth.magnitude, ">0", "Depth"],
                   [KMinor, ">=0", "K minor"])
    return (KMinor * FlowRate**2
            / (2 * u.gravity * (Width*Depth)**2)
            ).to(u.m)


@ut.list_handler()
def headloss_rect(FlowRate, Width, Depth, Length, KMinor, Nu, Roughness=None,
                  OpenChannel=None, *, PipeRough=None, openchannel=None):
    """Return the total head loss from major and minor losses in a rectangular
    channel.

    This equation applies to both laminar and turbulent flows.

    :param FlowRate: flow rate through channel
    :type FlowRate: u.m**3/u.s
    :param Width: width of channel
    :type Width: u.m
    :param Depth: depth of water in channel
    :type Depth: u.m
    :param Length: length of channel
    :type Length: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of channel
    :type Roughness: u.m
    :param OpenChannel: true if channel is open, false if closed
    :type OpenChannel: boolean

    :param PipeRough: deprecated; use Roughness instead
    :type openchannel: deprecated; use OpenChannel instead

    :return: total head loss in rectangular channel
    :rtype: u.m
    """
    if Roughness is not None and PipeRough is not None:
        raise TypeError("headloss_rect received both Roughness and PipeRough")
    elif Roughness is None and PipeRough is None:
        raise TypeError("headloss_rect missing Roughness argument")
    elif OpenChannel is not None and openchannel is not None:
        raise TypeError("headloss_rect received both OpenChannel and openchannel")
    elif OpenChannel is None and openchannel is None:
        raise TypeError("headloss_rect missing OpenChannel argument")
    else:
        if PipeRough is not None:
            warnings.warn("PipeRough is deprecated; use Roughness instead.",
                          UserWarning)
            Roughness = PipeRough
        if openchannel is not None:
            warnings.warn("openchannel is deprecated; use OpenChannel instead.",
                          UserWarning)
            OpenChannel = openchannel

    return (headloss_minor_rect(FlowRate, Width, Depth, KMinor)
            + headloss_major_rect(FlowRate, Width, Depth, Length,
                                  Nu, Roughness, OpenChannel))


@ut.list_handler()
def headloss_fric_general(Area, PerimWetted, Vel, Length, Nu, PipeRough):
    """
    .. deprecated::
        `headloss_fric_general` is deprecated; use `headloss_major_channel` instead.
    """
    warnings.warn('headloss_fric_general` is deprecated; use `headloss_major_channel` instead',
                  UserWarning)
    return headloss_major_channel(Area, PerimWetted, Vel, Length, Nu, PipeRough)


@ut.list_handler()
def headloss_major_channel(Area, PerimWetted, Vel, Length, Nu, Roughness):
    """Return the major head loss due to wall shear in a general channel.

    This equation applies to both laminar and turbulent flows.

    :param Area: cross sectional area of channel
    :type Area: u.m**2
    :param PerimWetted: wetted perimeter of channel
    :type PerimWetted: u.m
    :param Vel: velocity of fluid
    :type Vel: u.m/u.s
    :param Length: length of channel
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of channel
    :type Roughness: u.m

    :return: major head loss in general channel
    :rtype: u.m
    """
    ut.check_range([Length.magnitude, ">0", "Length"])
    return (fric_channel(Area, PerimWetted, Vel, Nu, Roughness) * Length
            / (4 * radius_hydraulic_channel(Area, PerimWetted))
            * Vel**2 / (2*u.gravity)
            ).to(u.m)


@ut.list_handler()
def headloss_exp_general(Vel, KMinor):
    """
    .. deprecated::
        `headloss_exp_general` is deprecated; use `headloss_minor_channel` instead.
    """
    warnings.warn('headloss_exp_general` is deprecated; use `headloss_minor_channel` instead',
                  UserWarning)
    return headloss_minor_channel(Vel, KMinor)


@ut.list_handler()
def headloss_minor_channel(Vel, KMinor):
    """Return the minor head loss due to expansion in a general channel.

    This equation applies to both laminar and turbulent flows.

    :param Vel: velocity of fluid
    :type Vel: u.m/u.s
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: minor head loss in general channel
    :rtype: u.m
    """
    ut.check_range([Vel.magnitude, ">0", "Velocity"],
                   [KMinor, '>=0', 'K minor'])
    return (KMinor * Vel**2 / (2*u.gravity)).to(u.m)


@ut.list_handler()
def headloss_gen(Area, Vel, PerimWetted, Length, KMinor, Nu, PipeRough):
    """
    .. deprecated::
        `headloss_gen` is deprecated; use `headloss_channel` instead.
    """
    warnings.warn('headloss_gen` is deprecated; use `headloss_channel` instead',
                  UserWarning)
    return headloss_channel(Area, Vel, PerimWetted, Length, KMinor, Nu, PipeRough)


@ut.list_handler()
def headloss_channel(Area, Vel, PerimWetted, Length, KMinor, Nu, Roughness):
    """Return the total head loss from major and minor losses in a general
    channel.

    This equation applies to both laminar and turbulent flows.

    :param Area: cross sectional area of channel
    :type Area: u.m**2
    :param Vel: velocity of fluid
    :type Vel: u.m/u.s
    :param PerimWetted: wetted perimeter of channel
    :type PerimWetted: u.m
    :param Length: length of channel
    :type Length: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of channel
    :type Roughness: u.m

    :return: total head loss in general channel
    :rtype: u.m
    """
    return (headloss_minor_channel(Vel, KMinor)
            + headloss_major_channel(Area, PerimWetted, Vel,
                                     Length, Nu, Roughness)).to(u.m)


@ut.list_handler()
def headloss_manifold(FlowRate, Diam, Length, KMinor, Nu, Roughness=None, NumOutlets=None, *, PipeRough=None):
    """Return the total head loss through the manifold.

    :param FlowRate: flow rate through manifold
    :type FlowRate: u.m**3/u.s
    :param Diam: diameter of manifold
    :type Diam: u.m
    :param Length: length of manifold
    :type Length: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of manifold
    :type Roughness: u.m
    :param NumOutlets: number of outlets from manifold
    :type NumOutlets: u.dimensionless or unitless

    :param PipeRough: deprecated; use Roughness instead

    :return: total headloss through manifold
    :rtype: u.m
    """
    ut.check_range([NumOutlets, ">0, int", 'Number of outlets'])

    if Roughness is not None and PipeRough is not None:
        raise TypeError("headloss_manifold received both Roughness and PipeRough")
    elif Roughness is None and PipeRough is None:
        raise TypeError("headloss_manifold missing Roughness argument")
    elif NumOutlets is None:
        raise TypeError("headloss_manifold missing NumOutlets argument")
    elif PipeRough is not None:
        warnings.warn("PipeRough is deprecated; use Roughness instead.",
                      UserWarning)
        Roughness = PipeRough

    return (headloss_pipe(FlowRate, Diam, Length, Nu, Roughness, KMinor)
            * ((1/3)
               + (1 / (2*NumOutlets))
               + (1 / (6*NumOutlets**2))
               )
            ).to(u.m)


@ut.list_handler()
def elbow_minor_loss(q, id_, k):
    """
    .. deprecated::
        `elbow_minor_loss` is deprecated; use `headloss_minor_elbow` instead.
    """
    warnings.warn('elbow_minor_loss is deprecated; use headloss_minor_elbow instead',
                  UserWarning)
    return headloss_minor_elbow(q, id_, k)


@ut.list_handler()
def headloss_minor_elbow(FlowRate, Diam, KMinor):
    """Return the minor head loss (due to changes in geometry) in an elbow.

    :param FlowRate: flow rate through pipe
    :type FlowRate: u.m**3/u.s
    :param Diam: diameter of pipe
    :type Diam: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: minor head loss in pipe
    :rtype: u.m
    """
    vel = FlowRate / area_circle(Diam)
    minor_loss = KMinor * vel ** 2 / (2 * u.gravity)
    return minor_loss.to(u.m)

######################### Orifices #########################


@ut.list_handler()
def flow_orifice(Diam, Height, RatioVCOrifice):
    """Return the flow rate of the orifice.

    :param Diam: diameter of orifice
    :type Diam: u.m
    :param Height: piezometric height of orifice
    :type Height: u.m
    :param RatioVCOrifice: vena contracta ratio of orifice
    :type RatioVCOrifice: u.dimensionless or unitless

    :return: flow rate of orifice
    :rtype: u.m**3/u.s
    """
    ut.check_range([Diam.magnitude, ">0", "Diameter"],
                   [RatioVCOrifice, "0-1", "VC orifice ratio"])
    if Height.magnitude > 0:
        return (RatioVCOrifice * area_circle(Diam)
                * np.sqrt(2 * u.gravity * Height)).to(u.m**3/u.s)
    else:
        return 0 * u.m**3/u.s


@ut.list_handler()
def flow_orifice_vert(Diam, Height, RatioVCOrifice):
    """Return the vertical flow rate of the orifice.

    :param Diam: diameter of orifice
    :type Diam: u.m
    :param Height: piezometric height of orifice
    :type Height: u.m
    :param RatioVCOrifice: vena contracta ratio of orifice
    :type RatioVCOrifice: u.dimensionless or unitless

    :return: vertical flow rate of orifice
    :rtype: u.m**3/u.s
    """
    ut.check_range([RatioVCOrifice, "0-1", "VC orifice ratio"])
    Diam = Diam.to(u.m)
    Height = Height.to(u.m)
    if Height > -Diam / 2:
        flow_vert = integrate.quad(lambda z: (Diam*np.sin(np.arccos(z*u.m/(Diam/2)))
                                              * np.sqrt(Height - z*u.m)
                                              ).magnitude,
                                   - Diam.magnitude / 2,
                                   min(Diam/2, Height).magnitude)
        return (flow_vert[0] * u.m**2.5 * RatioVCOrifice *
                np.sqrt(2 * u.gravity)).to(u.m**3/u.s)
    else:
        return 0 * u.m**3/u.s


@ut.list_handler()
def head_orifice(Diam, RatioVCOrifice, FlowRate):
    """Return the piezometric head of the orifice.

    :param Diam: diameter of orifice
    :type Diam: u.m
    :param RatioVCOrifice: vena contracta ratio of orifice
    :type RatioVCOrifice: u.dimensionless or unitless
    :param FlowRate: flow rate of orifice
    :type FlowRate: u.m**3/u.s

    :return: head of orifice
    :rtype: u.m
    """
    ut.check_range([Diam.magnitude, ">0", "Diameter"],
                   [FlowRate.magnitude, ">0", "Flow rate"],
                   [RatioVCOrifice, "0-1", "VC orifice ratio"])
    return ((FlowRate
             / (RatioVCOrifice * area_circle(Diam))
             )**2
            / (2*u.gravity)
            ).to(u.m)


@ut.list_handler()
def area_orifice(Height, RatioVCOrifice, FlowRate):
    """Return the area of the orifice.

    :param Height: piezometric height of orifice
    :type Height: u.m
    :param RatioVCOrifice: vena contracta ratio of orifice
    :type RatioVCOrifice: u.dimensionless or unitless
    :param FlowRate: flow rate of orifice
    :type FlowRate: u.m**3/u.s

    :return: area of orifice
    :rtype: u.m**2
    """
    ut.check_range([Height.magnitude, ">0", "Height"],
                   [FlowRate.magnitude, ">0", "Flow rate"],
                   [RatioVCOrifice, "0-1, >0", "VC orifice ratio"])
    return (FlowRate / (RatioVCOrifice * np.sqrt(2 * u.gravity *
                                                 Height))).to(u.m**2)


@ut.list_handler()
def num_orifices(FlowRate, RatioVCOrifice, HeadLossOrifice, DiamOrifice):
    """Return the number of orifices.

    :param FlowRate: flow rate of orifice
    :type FlowRate: u.m**3/u.s
    :param RatioVCOrifice: vena contracta ratio of orifice
    :type RatioVCOrifice: u.dimensionless or unitless
    :param HeadLossOrifice: head loss of orifice
    :type HeadLossOrifice: u.m
    :param DiamOrifice: diameter of orifice
    :type DiamOrifice: u.m

    :return: number of orifices
    :rtype: u.dimensionless
    """
    return np.ceil(area_orifice(HeadLossOrifice, RatioVCOrifice, FlowRate)
                   / area_circle(DiamOrifice)).to(u.dimensionless)

########################### Flows ###########################


@ut.list_handler()
def flow_transition(Diam, Nu):
    """Return the flow rate for the laminar/turbulent transition.

    :param Diam: diameter of pipe
    :type Diam: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s

    :return: flow rate for laminar/turbulent transition
    :rtype: u.m**3/u.s
    """
    ut.check_range([Diam.magnitude, ">0", "Diameter"],
                   [Nu.magnitude, ">0", "Nu"])
    return (np.pi * Diam * RE_TRANSITION_PIPE * Nu / 4).to(u.m**3/u.s)


@ut.list_handler()
def flow_hagen(Diam, HeadLossMajor=None, Length=None, Nu=None, *, HeadLossFric=None):
    """Return the flow rate for laminar flow with only major losses.

    :param Diam: diameter of pipe
    :type Diam: u.m
    :param HeadLossMajor: head loss due to friction
    :type HeadLossMajor: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s

    :param HeadLossFric: deprecated; use HeadLossMajor instead

    :return: flow rate for laminar flow with only major losses
    :rtype: u.m**3/u.s
    """
    if HeadLossMajor is not None and HeadLossFric is not None:
        raise TypeError("flow_hagen received both HeadLossMajor and HeadLossFric")
    elif HeadLossMajor is None and HeadLossFric is None:
        raise TypeError("flow_hagen missing HeadLossMajor argument")
    elif Length is None:
        raise TypeError("flow_hagen missing Length argument")
    elif Nu is None:
        raise TypeError("flow_hagen missing Nu argument")
    elif HeadLossFric is not None:
        warnings.warn("HeadLossFric is deprecated; use HeadLossMajor instead.",
                      UserWarning)
        HeadLossMajor = HeadLossFric

    ut.check_range([Diam.magnitude, ">0", "Diameter"],
                   [Length.magnitude, ">0", "Length"],
                   [HeadLossMajor.magnitude, ">=0", "Headloss due to friction"],
                   [Nu.magnitude, ">0", "Nu"])
    return ((np.pi*Diam**4) / (128*Nu) * u.gravity * HeadLossMajor
            / Length).to(u.m**3/u.s)


@ut.list_handler()
def flow_swamee(Diam, HeadLossMajor=None, Length=None, Nu=None, Roughness=None, *, HeadLossFric=None, PipeRough=None):
    """Return the flow rate for turbulent flow with only major losses.

    :param Diam: diameter of pipe
    :type Diam: u.m
    :param HeadLossMajor: head loss due to friction
    :type HeadLossMajor: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of pipe
    :type Roughness: u.m

    :param HeadLossFric: deprecated; use HeadLossMajor instead
    :param PipeRough: deprecated; use Roughness instead

    :return: flow rate for turbulent flow with only major losses
    :rtype: u.m**3/u.s
    """
    if HeadLossMajor is not None and HeadLossFric is not None:
        raise TypeError("flow_swamee received both HeadLossMajor and HeadLossFric")
    elif HeadLossMajor is None and HeadLossFric is None:
        raise TypeError("flow_swamee missing HeadLossMajor argument")
    elif Length is None:
        raise TypeError("flow_swamee missing Length argument")
    elif Nu is None:
        raise TypeError("flow_swamee missing Nu argument")
    elif Roughness is not None and PipeRough is not None:
        raise TypeError("flow_swamee received both Roughness and PipeRough")
    elif Roughness is None and PipeRough is None:
        raise TypeError("flow_swamee missing Roughness argument")
    else:
        if HeadLossFric is not None:
            warnings.warn("HeadLossFric is deprecated; use HeadLossMajor instead.",
                          UserWarning)
            HeadLossMajor = HeadLossFric
        if PipeRough is not None:
            warnings.warn("PipeRough is deprecated; use Roughness instead.",
                          UserWarning)
            Roughness = PipeRough

    ut.check_range([Diam.magnitude, ">0", "Diameter"],
                   [Length.magnitude, ">0", "Length"],
                   [HeadLossMajor.magnitude, ">0", "Headloss due to friction"],
                   [Nu.magnitude, ">0", "Nu"],
                   [Roughness.magnitude, ">=0", "Pipe roughness"])
    logterm = np.log10(Roughness / (3.7 * Diam)
                       + 2.51 * Nu * np.sqrt(Length / (2 * u.gravity
                                                       * HeadLossMajor
                                                       * Diam**3)
                                             )
                       )
    return ((-np.pi / np.sqrt(2)) * Diam**(5/2) * logterm
            * np.sqrt(u.gravity * HeadLossMajor / Length)
            ).to(u.m**3/u.s)


@ut.list_handler()
def flow_pipemajor(Diam, HeadLossFric, Length, Nu, PipeRough):
    """
    .. deprecated::
        `flow_pipemajor` is deprecated; use `flow_major_pipe` instead.
    """
    warnings.warn('flow_pipemajor is deprecated; use '
                  'flow_major_pipe instead.', UserWarning)
    return flow_major_pipe(Diam, HeadLossFric, Length, Nu, PipeRough)


@ut.list_handler()
def flow_major_pipe(Diam, HeadLossMajor, Length, Nu, Roughness):
    """Return the flow rate with only major losses.

    This function applies to both laminar and turbulent flows.

    :param Diam: diameter of pipe
    :type Diam: u.m
    :param HeadLossMajor: head loss due to friction
    :type HeadLossMajor: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of pipe
    :type Roughness: u.m

    :return: flow rate with only major losses
    :rtype: u.m**3/u.s
    """
    FlowHagen = flow_hagen(Diam, HeadLossMajor, Length, Nu)
    if FlowHagen < flow_transition(Diam, Nu):
        return FlowHagen
    else:
        return flow_swamee(Diam, HeadLossMajor, Length, Nu, Roughness)


@ut.list_handler()
def flow_pipeminor(Diam, HeadLossExpans, KMinor):
    """
    .. deprecated::
        `flow_pipeminor` is deprecated; use `flow_minor_pipe` instead.
    """
    warnings.warn('flow_pipeminor is deprecated; use '
                  'flow_minor_pipe instead.', UserWarning)
    return flow_minor_pipe(Diam, HeadLossExpans, KMinor)


@ut.list_handler()
def flow_minor_pipe(Diam, HeadLossMinor, KMinor):
    """Return the flow rate with only minor losses.

    This function applies to both laminar and turbulent flows.

    :param Diam: diameter of pipe
    :type Diam: u.m
    :param HeadLossExpans: head loss due to expansion
    :type HeadLossExpans: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: flow rate with only minor losses
    :rtype: u.m**3/u.s
    """
    ut.check_range([HeadLossMinor.magnitude, ">=0",
                    "Headloss due to expansion"],
                   [KMinor, ">0", "K minor"])
    return (area_circle(Diam) * np.sqrt(2 * u.gravity * HeadLossMinor
                                        / KMinor)
            ).to(u.m**3/u.s)


@ut.list_handler()
def flow_pipe(Diam, HeadLoss, Length, Nu, Roughness=None, KMinor=None, *, PipeRough=None):
    """Return the flow rate in a pipe.

    This function works for both major and minor losses as well as
    both laminar and turbulent flows.

    :param Diam: diameter of pipe
    :type Diam: u.m
    :param HeadLoss: total head loss from major and minor losses
    :type HeadLoss: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of pipe
    :type Roughness: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :param PipeRough: deprecated; use Roughness instead

    :return: flow rate in pipe
    :rtype: u.m**3/u.s
    """
    if Roughness is not None and PipeRough is not None:
        raise TypeError("flow_pipe received both Roughness and PipeRough")
    elif Roughness is None and PipeRough is None:
        raise TypeError("flow_pipe missing Roughness argument")
    elif KMinor is None:
        raise TypeError("flow_pipe missing KMinor argument")
    elif PipeRough is not None:
            warnings.warn("PipeRough is deprecated; use Roughness instead.",
                          UserWarning)
            Roughness = PipeRough

    if KMinor == 0:
        FlowRate = flow_major_pipe(Diam, HeadLoss, Length, Nu,
                                   Roughness)
    else:
        FlowRatePrev = 0
        err = 1.0
        FlowRate = min(flow_major_pipe(Diam, HeadLoss, Length,
                                       Nu, Roughness),
                       flow_minor_pipe(Diam, HeadLoss, KMinor)
                       )
        while err > 0.01:
            FlowRatePrev = FlowRate
            HLFricNew = (HeadLoss * headloss_major_pipe(FlowRate, Diam, Length,
                                                        Nu, Roughness)
                         / (headloss_major_pipe(FlowRate, Diam, Length,
                                                Nu, Roughness)
                            + headloss_minor_pipe(FlowRate, Diam, KMinor)
                            )
                         )
            FlowRate = flow_major_pipe(Diam, HLFricNew, Length,
                                       Nu, Roughness)
            if FlowRate == 0:
                err = 0.0
            else:
                err = (abs(FlowRate - FlowRatePrev)
                       / ((FlowRate + FlowRatePrev) / 2)
                       )
    return FlowRate.to(u.m**3/u.s)

########################## Diameters ##########################


@ut.list_handler()
def diam_hagen(FlowRate, HeadLossMajor=None, Length=None, Nu=None, *, HeadLossFric=None):
    """Return the inner diameter of a pipe with laminar flow and no minor losses.

    The Hagen Poiseuille equation is dimensionally correct and returns the
    inner diameter of a pipe given the flow rate and the head loss due
    to shear on the pipe walls. The Hagen Poiseuille equation does NOT take
    minor losses into account. This equation ONLY applies to laminar flow.

    :param FlowRate: flow rate of pipe
    :type FlowRate: u.m**3/u.s
    :param HeadLossFric: head loss due to friction
    :type HeadLossFric: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s

    :param HeadLossFric: deprecated; use HeadLossMajor instead

    :return: inner diameter of pipe
    :rtype: u.m
    """
    if HeadLossMajor is not None and HeadLossFric is not None:
        raise TypeError("diam_hagen received both HeadLossMajor and HeadLossFric")
    elif HeadLossMajor is None and HeadLossFric is None:
        raise TypeError("diam_hagen missing HeadLossMajor argument")
    elif Length is None:
        raise TypeError("diam_hagen missing Length argument")
    elif Nu is None:
        raise TypeError("diam_hagen missing Nu argument")
    elif HeadLossFric is not None:
        warnings.warn("HeadLossFric is deprecated; use HeadLossMajor instead.",
                      UserWarning)
        HeadLossMajor = HeadLossFric

    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Length.magnitude, ">0", "Length"],
                   [HeadLossMajor.magnitude, ">0", "Headloss due to friction"],
                   [Nu.magnitude, ">0", "Nu"])
    return (((128 * Nu * FlowRate * Length)
             / (u.gravity * HeadLossMajor * np.pi)
             ) ** (1/4)).to(u.m)


@ut.list_handler()
def diam_swamee(FlowRate, HeadLossMajor=None, Length=None, Nu=None, Roughness=None, *, HeadLossFric=None, PipeRough=None):
    """Return the inner diameter of a pipe with turbulent flow and no minor losses.

    The Swamee Jain equation is dimensionally correct and returns the
    inner diameter of a pipe given the flow rate and the head loss due
    to shear on the pipe walls. The Swamee Jain equation does NOT take
    minor losses into account. This equation ONLY applies to turbulent
    flow.

    :param FlowRate: flow rate of pipe
    :type FlowRate: u.m**3/u.s
    :param HeadLossFric: head loss due to friction
    :type HeadLossFric: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param PipeRough: roughness of pipe
    :type PipeRough: u.m

    :param HeadLossFric: deprecated; use HeadLossMajor instead
    :param PipeRough: deprecated; use Roughness instead

    :return: inner diameter of pipe
    :rtype: u.m
    """
    if HeadLossMajor is not None and HeadLossFric is not None:
        raise TypeError("diam_swamee received both HeadLossMajor and HeadLossFric")
    elif HeadLossMajor is None and HeadLossFric is None:
        raise TypeError("diam_swamee missing HeadLossMajor argument")
    elif Length is None:
        raise TypeError("diam_swamee missing Length argument")
    elif Nu is None:
        raise TypeError("diam_swamee missing Nu argument")
    elif Roughness is not None and PipeRough is not None:
        raise TypeError("diam_swamee received both Roughness and PipeRough")
    elif Roughness is None and PipeRough is None:
        raise TypeError("diam_swamee missing Roughness argument")
    else:
        if HeadLossFric is not None:
            warnings.warn("HeadLossFric is deprecated; use HeadLossMajor instead.",
                          UserWarning)
            HeadLossMajor = HeadLossFric
        if PipeRough is not None:
            warnings.warn("PipeRough is deprecated; use Roughness instead.",
                          UserWarning)
            Roughness = PipeRough

    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Length.magnitude, ">0", "Length"],
                   [HeadLossMajor.magnitude, ">0", "Headloss due to friction"],
                   [Nu.magnitude, ">0", "Nu"],
                   [Roughness.magnitude, ">=0", "Pipe roughness"])
    a = ((Roughness ** 1.25)
         * ((Length * FlowRate**2)
            / (u.gravity * HeadLossMajor)
            )**4.75
         ).to_base_units()
    b = (Nu**5 * FlowRate**47
         * (Length / (u.gravity * HeadLossMajor)) ** 26
         ).to_base_units()**0.2
    return (0.66 * (a+b)**0.04).to(u.m)


@ut.list_handler()
def diam_pipemajor(FlowRate, HeadLossFric, Length, Nu, PipeRough):
    """
    .. deprecated::
        `diam_pipemajor` is deprecated; use `diam_major_pipe` instead.
    """
    warnings.warn('diam_pipemajor is deprecated; use '
                  'diam_major_pipe instead.', UserWarning)
    return diam_major_pipe(FlowRate, HeadLossFric, Length, Nu, PipeRough)


@ut.list_handler()
def diam_major_pipe(FlowRate, HeadLossMajor, Length, Nu, Roughness):
    """Return the pipe inner diameter that would result in given major losses.

    This function applies to both laminar and turbulent flow.

    :param FlowRate: flow rate of pipe
    :type FlowRate: u.m**3/u.s
    :param HeadLossMajor: head loss due to friction
    :type HeadLossMajor: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param Roughness: roughness of pipe
    :type Roughness: u.m

    :return: inner diameter of pipe
    :rtype: u.m
    """
    DiamLaminar = diam_hagen(FlowRate, HeadLossMajor, Length, Nu)
    if re_pipe(FlowRate, DiamLaminar, Nu) <= RE_TRANSITION_PIPE:
        return DiamLaminar
    else:
        return diam_swamee(FlowRate, HeadLossMajor, Length,
                           Nu, Roughness)


@ut.list_handler()
def diam_pipeminor(FlowRate, HeadLossExpans, KMinor):
    """
    .. deprecated::
        `diam_pipeminor` is deprecated; use `diam_minor_pipe` instead.
    """
    warnings.warn('diam_pipeminor is deprecated; use '
                  'diam_minor_pipe instead.', UserWarning)
    return diam_minor_pipe(FlowRate, HeadLossExpans, KMinor)


@ut.list_handler()
def diam_minor_pipe(FlowRate, HeadLossMinor, KMinor):
    """Return the pipe inner diameter that would result in the given minor losses.

    This function applies to both laminar and turbulent flow.

    :param FlowRate: flow rate of pipe
    :type FlowRate: u.m**3/u.s
    :param HeadLossMinor: head loss due to expansion
    :type HeadLossMinor: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: inner diameter of pipe
    :rtype: u.m
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [KMinor, ">=0", "K minor"],
                   [HeadLossMinor.magnitude, ">0", "Headloss due to expansion"])
    return (np.sqrt(4 * FlowRate / np.pi)
            * (KMinor / (2 * u.gravity * HeadLossMinor)) ** (1/4)
            ).to(u.m)


@ut.list_handler()
def diam_pipe(FlowRate, HeadLoss, Length, Nu, PipeRough, KMinor):
    """Return the pipe inner diameter that would result in the given total head
    loss.

    This function applies to both laminar and turbulent flow and
    incorporates both minor and major losses.

    :param FlowRate: flow rate of pipe
    :type FlowRate: u.m**3/u.s
    :param HeadLoss: total head loss from major and minor losses
    :type HeadLoss: u.m
    :param Length: length of pipe
    :type Length: u.m
    :param Nu: kinematic viscosity of fluid
    :type Nu: u.m**2/u.s
    :param PipeRough: roughness of pipe
    :type PipeRough: u.m
    :param KMinor: minor loss coefficient
    :type KMinor: u.dimensionless or unitless

    :return: inner diameter of pipe
    :rtype: u.m
    """
    if KMinor == 0:
        Diam = diam_major_pipe(FlowRate, HeadLoss, Length, Nu,
                               PipeRough)
    else:
        Diam = max(diam_major_pipe(FlowRate, HeadLoss,
                                   Length, Nu, PipeRough),
                   diam_minor_pipe(FlowRate, HeadLoss, KMinor))
        err = 1.00
        while err > 0.001:
            DiamPrev = Diam
            HLFricNew = (HeadLoss * headloss_major_pipe(FlowRate, Diam, Length,
                                                        Nu, PipeRough
                                                        )
                         / (headloss_major_pipe(FlowRate, Diam, Length,
                                                Nu, PipeRough
                                                )
                            + headloss_minor_pipe(FlowRate, Diam, KMinor
                                                  )
                            )
                         )
            Diam = diam_major_pipe(FlowRate, HLFricNew, Length, Nu, PipeRough
                                   )
            err = abs(Diam - DiamPrev) / ((Diam + DiamPrev) / 2)
    return Diam.to(u.m)


@ut.list_handler()
def pipe_ID(FlowRate, Pressure):
    """Return the inner diameter of a pipe for a given pressure
    recovery constraint.

    :param FlowRate: flow rate of pipe
    :type FlowRate: u.m**3/u.s
    :param Pressure: pressure recovery constraint
    :type Pressure: u.m

    :return: inner diameter of pipe
    :rtype: u.m
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Pressure.magnitude, ">0", "Pressure"])
    return np.sqrt(FlowRate/((np.pi/4)*np.sqrt(2*u.gravity*Pressure))).to(u.m)

############################ Weirs ############################


@ut.list_handler()
def width_rect_weir(FlowRate, Height):
    """
    .. deprecated::
        `width_rect_weir` is deprecated; use `width_weir_rect` instead.
    """
    warnings.warn('width_rect_weir is deprecated; use '
                  'width_weir_rect instead.', UserWarning)
    return width_weir_rect(FlowRate, Height)


@ut.list_handler()
def width_weir_rect(FlowRate, Height):
    """Return the width of a rectangular weir given its flow rate and the
    height of the water above the weir. For a weir that is a vertical pipe,
    this value is the circumference.

    :param FlowRate: flow rate over weir
    :type FlowRate: u.m**3/u.s
    :param Height: height of water above weir
    :type Height: u.m

    :return: width of weir
    :rtypes: u.m
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Height.magnitude, ">0", "Height"])
    return ((3 / 2) * FlowRate / (con.VC_ORIFICE_RATIO
                                  * np.sqrt(2 * u.gravity) * Height ** (3 / 2))
            ).to(u.m)


@ut.list_handler()
def headloss_weir(FlowRate, Width):
    """
    .. deprecated::
        `headloss_weir` is deprecated; use `headloss_weir_rect` instead.
    """
    warnings.warn('headloss_weir is deprecated; use '
                  'headloss_weir_rect instead.', UserWarning)
    return headloss_weir_rect(FlowRate, Width)


@ut.list_handler()
def headloss_weir_rect(FlowRate, Width):
    """Return the head loss of a rectangular or vertical pipe weir.

    Head loss for a weir is the difference in height between the water
    upstream of the weir and the top of the weir.

    :param FlowRate: flow rate over weir
    :type FlowRate: u.m**3/u.s
    :param Width: width of weir (circumference for a vertical pipe)
    :type Width: u.m

    :return: head loss of weir
    :rtypes: u.m
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Width.magnitude, ">0", "Width"])
    return ((((3/2) * FlowRate
              / (con.VC_ORIFICE_RATIO * np.sqrt(2 * u.gravity) * Width)
              ) ** 2).to(u.m**3)) ** (1/3)


@ut.list_handler()
def flow_rect_weir(Height, Width):
    """
    .. deprecated::
        `flow_rect_weir` is deprecated; use `flow_weir_rect` instead.
    """
    warnings.warn('flow_rect_weir is deprecated; use '
                  'flow_weir_rect instead.', UserWarning)
    return flow_weir_rect(Height, Width)


@ut.list_handler()
def flow_weir_rect(Height, Width):
    """Return the flow rate of a rectangular or vertical pipe weir.

    :param Height: height of water above weir
    :type Height: u.m
    :param Width: width of weir (circumference for a vertical pipe)
    :type Width: u.m

    :return: flow of weir
    :rtype: u.m**3/u.s
    """
    ut.check_range([Height.magnitude, ">0", "Height"],
                   [Width.magnitude, ">0", "Width"])
    return ((2/3) * con.VC_ORIFICE_RATIO
            * (np.sqrt(2*u.gravity) * Height**(3/2))
            * Width).to(u.m**3/u.s)

######################## Porous Media ########################

class DeprecatedFunctionError(Exception):
    def __init__(self, message):
        self.message = message

@ut.list_handler()
def headloss_kozeny(Length, DiamMedia=None, ApproachVel=None, Porosity=None, Nu=None, *, Diam=None, Vel=None):
    """
    .. deprecated::
        `headloss_kozeny` is deprecated; use `headloss_ergun` instead.
    """
    raise DeprecatedFunctionError("This function is deprecated. Please use headloss_ergun.")


@ut.list_handler()
def re_ergun(ApproachVel, DiamMedia, Temperature, Porosity):
    """Return the Reynolds number for flow through porous media.

    :param ApproachVel: approach velocity or superficial fluid velocity
    :type ApproachVel: u.m/u.s
    :param DiamMedia: particle diameter
    :type DiamMedia: u.m
    :param Temperature: temperature of porous medium
    :type Temperature: u.degK
    :param Porosity: porosity of porous medium
    :type Porosity: u.dimensionless or unitless

    :return: Reynolds number for flow through porous media
    :rtype: u.dimensionless
    """
    ut.check_range([ApproachVel.magnitude, ">0", "ApproachVel"],
                   [DiamMedia.magnitude, ">0", "DiamMedia"],
                   [Porosity, "0-1", "Porosity"])
    if Porosity == 1:
        raise ValueError("Porosity is " + str(Porosity) + " must be great than\
            or equal to 0 and less than 1")
    return (ApproachVel * DiamMedia /
            (viscosity_kinematic_water(Temperature)
             * (1 - Porosity))).to(u.dimensionless)


@ut.list_handler()
def fric_ergun(ApproachVel, DiamMedia, Temperature, Porosity):
    """Return the friction factor for flow through porous media.

    :param ApproachVel: superficial fluid velocity (VelSuperficial?)
    :type ApproachVel: u.m/u.s
    :param DiamMedia: particle diameter
    :type DiamMedia: u.m
    :param Temperature: temperature of porous medium
    :type Temperature: u.degK
    :param Porosity: porosity of porous medium
    :type Porosity: u.dimensionless or unitless

    :return: friction factor for flow through porous media
    :rtype: u.dimensionless
    """
    return (300 / re_ergun(ApproachVel, DiamMedia, Temperature, Porosity)
            + 3.5 * u.dimensionless)


@ut.list_handler()
def headloss_ergun(ApproachVel, DiamMedia, Temperature, Porosity, Length):
    """Return the frictional head loss for flow through porous media.

    :param ApproachVel: superficial fluid velocity (VelSuperficial?)
    :type ApproachVel: u.m/u.s
    :param DiamMedia: particle diameter
    :type DiamMedia: u.m
    :param Temperature: temperature of porous medium
    :type Temperature: u.degK
    :param Porosity: porosity of porous medium
    :type Porosity: u.dimensionless or unitless
    :param Length: length of pipe or duct
    :type Length: u.m

    :return: frictional head loss for flow through porous media
    :rtype: u.m
    """
    return (fric_ergun(ApproachVel, DiamMedia, Temperature, Porosity)
            * Length / DiamMedia * ApproachVel**2 / (2*u.gravity) * (1-Porosity)
            / Porosity**3).to(u.m)


@ut.list_handler()
def g_cs_ergun(ApproachVel, DiamMedia, Temperature, Porosity):
    """Camp Stein velocity gradient for flow through porous media.

    :param ApproachVel: superficial fluid velocity (VelSuperficial?)
    :type ApproachVel: u.m/u.s
    :param DiamMedia: particle diameter
    :type DiamMedia: u.m
    :param Temperature: temperature of porous medium
    :type Temperature: u.degK
    :param Porosity: porosity of porous medium
    :type Porosity: u.dimensionless or unitless

    :return: Camp Stein velocity gradient for flow through porous media
    :rtype: u.Hz
    """
    return np.sqrt(fric_ergun(ApproachVel, DiamMedia, Temperature, Porosity)
                   * ApproachVel**3 * (1-Porosity)
                   / (2 * viscosity_kinematic_water(Temperature) * DiamMedia
                      * Porosity**4)).to(u.Hz)

######################## Miscellaneous ########################


@ut.list_handler()
def height_water_critical(FlowRate, Width):
    """Return the critical local water height.

    :param FlowRate: flow rate of water
    :type FlowRate: u.m**3/u.s
    :param Width: width of channel (????????)
    :type Width: u.m

    :return: critical water height
    :rtype: u.m
    """
    ut.check_range([FlowRate.magnitude, ">0", "Flow rate"],
                   [Width.magnitude, ">0", "Width"])
    return ((FlowRate / (Width * np.sqrt(1*u.gravity))) ** (2/3)).to(u.m)


@ut.list_handler()
def vel_horizontal(HeightWaterCritical):
    """Return the horizontal velocity. (at the critical water depth??????)

    :param HeightWaterCritical: critical water height
    :type HeightWaterCritical: u.m

    :return: horizontal velocity
    :rtype: u.m/u.s
    """
    ut.check_range([HeightWaterCritical.magnitude, ">0", "Critical height of water"])
    return np.sqrt(u.gravity * HeightWaterCritical).to(u.m/u.s)


@ut.list_handler()
def manifold_id_alt(q, pr_max):
    """Return the inner diameter of a manifold when major losses are
    negligible.
    """
    manifold_id_alt = np.sqrt(
        4 * q / (
            np.pi * np.sqrt(
                2 * u.gravity * pr_max
            )
        )
    )
    return manifold_id_alt


@ut.list_handler()
def manifold_id(q, h, l, q_ratio, nu, eps, k, n):
    id_new = 2 * u.inch
    id_old = 0 * u.inch
    error = 1
    while error > 0.01:
        id_old = id_new
        id_new = (
            ((8 * q ** 2) / (u.gravity * np.pi ** 2 * h)) *
            (
                (
                    1 + fric_pipe(q, id_old, nu, eps) *
                    (1 / 3 + 1 / (2 * n) + 1 / (6 * n ** 2))
                ) /
                (1 - q_ratio ** 2)
            )
        ) ** (1 / 4)
        error = np.abs(id_old - id_new) / id_new
    return id_new


@ut.list_handler()
def manifold_nd(q, h, l, q_ratio, nu, eps, k, n, sdr):
    manifold_nd = pipe.ND_SDR_available(
            manifold_id(q, h, l, q_ratio, nu, eps, k, n),
            sdr
        )
    return manifold_nd


@ut.list_handler()
def horiz_chan_w(q, depth, hl, l, nu, eps, manifold, k):
    hl = min(hl, depth / 3)
    horiz_chan_w_new = q / ((depth - hl) * np.sqrt(2 * u.gravity * hl))

    error = 1
    i = 0
    while error > 0.001 and i < 20:
        w = horiz_chan_w_new
        i = i + 1
        horiz_chan_w_new = np.sqrt(
            (
                1 + k +
                    fric_rect(q, w, depth - hl, nu, eps, True) *
                    (l / (4 * radius_hydraulic_rect(w, depth - hl, True))) *
                    (1 - (2 * (int(manifold) / 3)))
            ) / (2 * u.gravity * hl)
        ) * (q / (depth - hl))
        error = np.abs(horiz_chan_w_new - w) / (horiz_chan_w_new + w)
    return horiz_chan_w_new.to(u.m)


@ut.list_handler()
def horiz_chan_h(q, w, hl, l, nu, eps, manifold):
    h_new = (q / (w * np.sqrt(2 * u.gravity * hl))) + hl
    error = 1
    i = 0
    while error > 0.001 and i < 200:
        h = h_new
        hl_local = min(hl, h / 3)
        i = i + 1
        h_new = (q/ w) * np.sqrt((1 + \
            fric_rect(q, w, h - hl_local, nu, eps, True) * (l / (4 * \
                radius_hydraulic_rect(w, h - hl_local, True))) * (1 - 2 * (int(manifold) / 3))
        )/ (2 * u.gravity * hl_local)) + (hl_local)
        error = np.abs(h_new - h) / (h_new + h)
    return h_new.to(u.m)


@ut.list_handler()
def pipe_flow_nd(q, sdr, hl, l, nu, eps, k):
    i = 0
    id_sdr_all_available = pipe.ID_SDR_all_available(sdr)
    while q > flow_pipe(id_sdr_all_available[i], hl, l, nu, eps, k):
        i_d = id_sdr_all_available[i]
        i += 1
    return pipe.ND_SDR_available(i_d, sdr)



from aguaclara.core.constants import *
from aguaclara.core.drills import *
from aguaclara.core.head_loss import *
from aguaclara.core.materials import *
from aguaclara.core.physchem import *
from aguaclara.core.pipes import *
from aguaclara.core.units import *
from aguaclara.core.utility import *
from aguaclara.core.onshape_parser import *

from aguaclara.design.cdc import CDC
from aguaclara.design.component import Component
from aguaclara.design.ent_floc import EntTankFloc
from aguaclara.design.ent import EntranceTank
from aguaclara.design.filter import Filter
from aguaclara.design.floc import Flocculator
import aguaclara.design.human_access as ha
from aguaclara.design.lfom import LFOM
from aguaclara.design.plant import Plant
from aguaclara.design.sed_chan import SedimentationChannel
from aguaclara.design.sed_tank import SedimentationTank
from aguaclara.design.sed import Sedimentor

from aguaclara.research.environmental_processes_analysis import *
from aguaclara.research.floc_model import *
from aguaclara.research.procoda_parser import *
from aguaclara.research.peristaltic_pump import *
from aguaclara.research.stock_qc import *



# -*- coding: utf-8 -*-
"""This file contains functions which can be used to model the behavior of
flocs based on the chemical interactions of clay, coagulant, and humic acid.
"""

######################### Imports #########################
import numpy as np
from aguaclara.core.units import u
from aguaclara.core import physchem as pc, utility as ut

u.enable_contexts('chem')

##################### Class Definition #####################


class Material:
    """A particulate material with a name, diameter, density, and
    molecular weight.
    """

    def __init__(self, name, diameter, density, molecWeight):
        """Initialize a Material object.

        :param name: Name of the material
        :type name: string
        :param diameter: Diameter of the material in particulate form
        :type diameter: float
        :param density: Density of the material (mass/volume)
        :type density: float
        :param molecWeight: Molecular weight of the material (mass/mole)
        :type molecWeight: float
        """
        self.name = name
        self.Diameter = diameter
        self.Density = density
        self.MolecWeight = molecWeight


class Chemical(Material):
    """A chemical with a name, diameter, density, molecular weight, number of
    aluminum atoms per molecule, and a precipitate.
    """

    def __init__(self, name, diameter, density, molecWeight, Precipitate,
                 AluminumMPM=None):
        """Initialize a Chemical object.

        :param name: Name of the material
        :type name: string
        :param diameter: Diameter of the material in particulate form
        :type diameter: length
        :param density: Density of the material
        :type density: mass/length**3
        :param molecWeight: Molecular weight of the material
        :type molecWeight: mass/mole
        :param Precipitate: Name of the precipitate
        :type Precipitate: string
        :param AluminumMPM: aluminum atoms per molecule
        :type AluminumMPM: int
        """
        Material.__init__(self, name, diameter, density, molecWeight)
        self.AluminumMPM = AluminumMPM
        self.Precip = Precipitate
        if self.Precip == self.name:
            self.PrecipName = name
            self.PrecipDiameter = diameter
            self.PrecipDensity = density
            self.PrecipMolecWeight = molecWeight
            self.PrecipAluminumMPM = AluminumMPM
        else:
            self.PrecipName = Precipitate

    def define_Precip(self, diameter, density, molecweight, alumMPM):
        """Define a precipitate for the chemical.

        :param diameter: Diameter of the precipitate in particulate form
        :type diameter: float
        :param density: Density of the material (mass/volume)
        :type density: float
        :param molecWeight: Molecular weight of the material (mass/mole)
        :type molecWeight: float
        :param alumMPM: aluminum atoms per molecule
        :type alumMPM: int
        """
        self.PrecipDiameter = diameter
        self.PrecipDensity = density
        self.PrecipMolecWeight = molecweight
        self.PrecipAluminumMPM = alumMPM


################## Material Definitions ##################

#: A Material representing clay with a diameter of 7e-6 m and density of 2650
#: kg/m^3.
Clay = Material('Clay', 7 * 10**-6 * u.m, 2650 * u.kg/u.m**3, None)

#: A Material representing polyaluminum chloride (PACl) with a diameter of
#: 9e-8m, density of 1138 kg/m^2, and molecular weight of 1.039 kg/mol. It is
#: its own precipitate.
PACl = Chemical('PACl', 9 * 10 **-8 * u.m, 1138 * u.kg/u.m**3,
                1.039 * u.kg/u.mol, 'PACl', AluminumMPM=13)

#: A Material representing alum with a diameter of 7e-8 m, density of
#: 2420 kg/m^3, and molecular weight of 0.59921 kg/mol. It's precipitate is
#: AlOH3, with the same diameter and density, and a molecular weight of 0.078
#: kg/mol.
Alum = Chemical('Alum', 7 * 10 **-8 * u.m, 2420 * u.kg/u.m**3,
                0.59921 * u.kg/u.mol, 'AlOH3', AluminumMPM=2)
Alum.define_Precip(7 * 10 **-8 * u.m, 2420 * u.kg/u.m**3,
                   0.078 * u.kg/u.mol, 1)

#: A Material representing humic acid with a diameter of 7.2e-8 m and density
#: of 1780 kg/m^3. It is its own precipitate.
HumicAcid = Chemical('Humic Acid', 72 * 10**-9 * u.m, 1780 * u.kg/u.m**3, None,
                     'Humic Acid')


################### Necessary Constants ###################
#: Fractal dimension, based on data from published in Environmental Engineering
#: Science, "Fractal Models for Floc Density, Sedimentation Velocity, and Floc
#: Volume Fraction for High Peclet Number Reactors" by Monroe Weber-Shirk and
#: Leonard Lion (2015).
DIM_FRACTAL = 2.3

#: Ratio of clay platelet height to diameter.
RATIO_HEIGHT_DIAM = 0.1

#: Ratio between inner viscous length scale and Kolmogorov length scale.
RATIO_KOLMOGOROV = 50

#: Shape factor for drag on flocs used in terminal velocity equation.
PHI_FLOC = 45/24

#: The Avogadro constant.
NUM_AVOGADRO = 6.0221415 * 10**23

#: Molecular weight of aluminum in kg/mole.
MOLEC_WEIGHT_ALUMINUM = 0.027 * u.kg / u.mol


######################## Functions ########################
# @u.wraps(u.kg/u.m**3, None, False)
@ut.list_handler()
def dens_alum_nanocluster(coag):
    """Return the density of the aluminum in the nanocluster.

    This is useful for determining the volume of nanoclusters
    given a concentration of aluminum.
    """
    density = (coag.PrecipDensity * MOLEC_WEIGHT_ALUMINUM
               * coag.PrecipAluminumMPM / coag.PrecipMolecWeight)
    return density.to(u.kg/u.m**3)


# @u.wraps(u.kg/u.m**3, [u.kg/u.m**3, u.degK], False)
@ut.list_handler()
def dens_pacl_solution(ConcAluminum, temp):
    """Return the density of the PACl solution.

    From Stock Tank Mixing report Fall 2013:
    https://confluence.cornell.edu/download/attachments/137953883/20131213_Research_Report.pdf
    """
    return ((0.492 * ConcAluminum * PACl.MolecWeight
             / (PACl.AluminumMPM * MOLEC_WEIGHT_ALUMINUM)
             ) + pc.density_water(temp)
            ).to(u.kg/u.m**3)


@ut.list_handler()
def conc_precipitate(ConcAluminum, coag):
    """Return coagulant precipitate concentration given aluminum dose.

    This function assumes complete precipitation of coagulant into Al13.

    Note that conc_precipitate returns a value that varies from the equivalent
    MathCAD function beginning at the third decimal place. The majority of
    functions below this point in the file ultimately call on conc_precipitate
    at some point, and will not return the same value as their equivalent
    function in MathCAD. This is known.

    :param ConcAluminum: Concentration of aluminum in solution
    :type ConcAluminum: float
    :param coag: Type of coagulant in solution, e.g. floc_model.PACl
    :type coag: floc_model.Material

    :return: Concentration of coagulant precipitates
    :rtype: float
    """
    return ((ConcAluminum / MOLEC_WEIGHT_ALUMINUM)
            * (coag.PrecipMolecWeight / coag.PrecipAluminumMPM)
            ).to(u.kg/u.m**3)


# @u.wraps(u.kg/u.m**3, [u.kg/u.m**3, u.kg/u.m**3, None], False)
@ut.list_handler()
def conc_floc(ConcAluminum, concClay, coag):
    """Return floc density given aluminum dose, turbidity, and coagulant"""
    return (conc_precipitate(ConcAluminum, coag) + concClay).to(u.kg/u.m**3)


# @u.wraps(u.mol/u.m**3, u.kg/u.m**3, False)
@ut.list_handler()
def moles_aluminum(ConcAluminum):
    """Return the # of moles aluminum given aluminum concentration."""
    return (ConcAluminum / MOLEC_WEIGHT_ALUMINUM).to(u.mol/u.m**3)


# @u.wraps(u.m, u.kg/u.m**3, False)
@ut.list_handler()
def sep_dist_aluminum(ConcAluminum):
    """Return the separation distance between aluminum molecules."""
    return ((1 / (NUM_AVOGADRO / u.mol * moles_aluminum(ConcAluminum)))**(1/3)).to(u.m)


@ut.list_handler()
def particle_number_concentration(ConcMat, material):
    """Return the number of particles in suspension.

    :param ConcMat: Concentration of the material
    :type ConcMat: float
    :param material: The material in solution
    :type material: floc_model.Material
    """
    return (ConcMat / ((material.Density * np.pi * material.Diameter**3) / 6)).to(u.m**-3)


# @u.wraps(u.m, [u.kg/u.m**3, u.m], False)
@ut.list_handler()
def sep_dist_clay(ConcClay, material):
    """Return the separation distance between clay particles."""
    return (((material.Density/ConcClay)
             * ((np.pi * material.Diameter ** 3) / 6)) ** (1 / 3)).to(u.m)


# @u.wraps(1/u.m**3, [u.kg/u.m**3, None], False)
@ut.list_handler()
def num_nanoclusters(ConcAluminum, coag):
    """Return the number of Aluminum nanoclusters."""
    return (ConcAluminum / (dens_alum_nanocluster(coag)
                            * np.pi * coag.Diameter**3)).to(1/u.m**3)


@ut.list_handler()
def frac_vol_floc_initial(ConcAluminum, ConcClay, coag, material):
    """Return the volume fraction of flocs initially present, accounting for both suspended particles and coagulant precipitates.

    :param ConcAluminum: Concentration of aluminum in solution
    :type ConcAluminum: float
    :param ConcClay: Concentration of particle in suspension
    :type ConcClay: float
    :param coag: Type of coagulant in solution
    :type coag: float
    :param material: Type of particles in suspension, e.g. floc_model.Clay
    :type material: floc_model.Material

    :return: Volume fraction of particles initially present
    :rtype: float
    """
    return ((conc_precipitate(ConcAluminum, coag)/coag.PrecipDensity)
            + (ConcClay / material.Density)).to(u.dimensionless)


####################### p functions #######################
@ut.list_handler()
def p(C, C0=1):
    return -np.log10(C/C0)


@ut.list_handler()
def invp(pC, C0=1):
    return C0 * 10**-pC


#################### Fractal functions ####################
# @u.wraps(u.m, [u.dimensionless, u.m, u.dimensionless], False)
@ut.list_handler()
def diam_fractal(DIM_FRACTAL, DiamInitial, NumCol):
    """Return the diameter of a floc given NumCol doubling collisions."""
    return (DiamInitial * 2**(NumCol / DIM_FRACTAL)).to(u.m)


# @u.wraps(None, [u.dimensionless, None, u.m], False)
@ut.list_handler()
def num_coll_reqd(DIM_FRACTAL, material, DiamTarget):
    """Return the number of doubling collisions required.

    Calculates the number of doubling collisions required to produce
    a floc of diameter DiamTarget.
    """
    return (DIM_FRACTAL * np.log(DiamTarget/material.Diameter)/np.log(2)).to(u.dimensionless)


# @u.wraps(u.m, [u.kg/u.m**3, u.kg/u.m**3, None, None,
               # u.dimensionless, u.m], False)
@ut.list_handler()
def sep_dist_floc(ConcAluminum, ConcClay, coag, material,
                  DIM_FRACTAL, DiamTarget):
    """Return separation distance as a function of floc size."""
    return (material.Diameter
            * (np.pi/(6
                      * frac_vol_floc_initial(ConcAluminum, ConcClay,
                                              coag, material)
                      ))**(1/3)
            * (DiamTarget / material.Diameter)**(DIM_FRACTAL / 3)
            ).to(u.m)


# @u.wraps(u.m, [u.kg/u.m**3, u.kg/u.m**3, None, u.dimensionless,
#                None, u.m], False)
@ut.list_handler()
def frac_vol_floc(ConcAluminum, ConcClay, coag, DIM_FRACTAL,
                  material, DiamTarget):
    """Return the floc volume fraction."""
    return (frac_vol_floc_initial(ConcAluminum, ConcClay, coag, material)
            * (DiamTarget / material.Diameter)**(3-DIM_FRACTAL)
            ).to(u.dimensionless)


# @u.wraps(u.kg/u.m**3, [u.kg/u.m**3, u.kg/u.m**3, None, None], False)
@ut.list_handler()
def dens_floc_init(ConcAluminum, ConcClay, coag, material):
    """Return the density of the initial floc.

    Initial floc is made primarily of the primary colloid and nanoglobs.
    """
    return (conc_floc(ConcAluminum, ConcClay, coag)
            / frac_vol_floc_initial(ConcAluminum, ConcClay, coag, material)
            ).to(u.kg/u.m**3)


#################### Flocculation Model ####################
# @u.wraps(None, u.m, False)
@ut.list_handler()
def ratio_clay_sphere(RatioHeightDiameter):
    """Return the surface area to volume ratio for clay.

    Normalized by surface area to volume ratio for a sphere.
    """
    return ((1/2 + RatioHeightDiameter) * (2 / (3*RatioHeightDiameter))**(2/3))*u.dimensionless


@ut.list_handler()
def ratio_area_clay_total(ConcClay, material, DiamTube, RatioHeightDiameter):
    """Return the surface area of clay normalized by total surface area.

    Total surface area is a combination of clay and reactor wall
    surface areas. This function is used to estimate how much coagulant
    actually goes to the clay.

    :param ConcClay: Concentration of clay in suspension
    :type ConcClay: float
    :param material: Type of clay in suspension, e.g. floc_model.Clay
    :type material: floc_model.Material
    :param DiamTube: Diameter of flocculator tube (assumes tube flocculator for calculation of reactor surface area)
    :type DiamTube: float
    :param RatioHeightDiameter: Dimensionless ratio describing ratio of clay height to clay diameter
    :type RatioHeightDiameter: float

    :return: The ratio of clay surface area to total available surface area (accounting for reactor walls)
    :rtype: float
    """
    return (1
            / (1
               + (2 * material.Diameter
                  / (3 * DiamTube * ratio_clay_sphere(RatioHeightDiameter)
                     * (ConcClay / material.Density)
                     )
                  )
               )
            ).to(u.dimensionless)


@ut.list_handler()
def gamma_coag(ConcClay, ConcAluminum, coag, material,
               DiamTube, RatioHeightDiameter):
    """Return the coverage of clay with nanoglobs.

    This function accounts for loss to the tube flocculator walls
    and a poisson distribution on the clay given random hits by the
    nanoglobs. The poisson distribution results in the coverage only
    gradually approaching full coverage as coagulant dose increases.

    :param ConcClay: Concentration of clay in suspension
    :type ConcClay: float
    :param ConcAluminum: Concentration of aluminum in solution
    :type ConcAluminum: float
    :param coag: Type of coagulant in solution, e.g. floc_model.PACl
    :type coag: floc_model.Material
    :param material: Type of clay in suspension, e.g. floc_model.Clay
    :type material: floc_model.Material
    :param DiamTube: Diameter of flocculator tube (assumes tube flocculator for calculation of reactor surface area)
    :type DiamTube: float
    :param RatioHeightDiameter: Dimensionless ratio of clay height to clay diameter
    :type RatioHeightDiameter: float

    :return: Fraction of the clay surface area that is coated with coagulant precipitates
    :rtype: float
    """
    return (1 - np.exp((
                       (-frac_vol_floc_initial(ConcAluminum, 0*u.kg/u.m**3, coag, material)
                        * material.Diameter)
                        / (frac_vol_floc_initial(0*u.kg/u.m**3, ConcClay, coag, material)
                           * coag.Diameter))
                       * (1 / np.pi)
                       * (ratio_area_clay_total(ConcClay, material,
                                                DiamTube, RatioHeightDiameter)
                          / ratio_clay_sphere(RatioHeightDiameter))
                       )).to(u.dimensionless)


# @u.wraps(None, [u.kg/u.m**3, u.kg/u.m**3, None, None], False)
@ut.list_handler()
def gamma_humic_acid_to_coag(ConcAl, ConcNatOrgMat, NatOrgMat, coag):
    """Return the fraction of the coagulant that is coated with humic acid.

    :param ConcAl: Concentration of alumninum in solution
    :type ConcAl: float
    :param ConcNatOrgMat: Concentration of natural organic matter in solution
    :type ConcNatOrgMat: float
    :param NatOrgMat: type of natural organic matter, e.g. floc_model.HumicAcid
    :type NatOrgMat: floc_model.Material
    :param coag: Type of coagulant in solution, e.g. floc_model.PACl
    :type coag: floc_model.Material

    :return: fraction of the coagulant that is coated with humic acid
    :rtype: float
    """
    return (min(((ConcNatOrgMat / conc_precipitate(ConcAl, coag))
                 * (coag.Density / NatOrgMat.Density)
                 * (coag.Diameter / (4 * NatOrgMat.Diameter))
                 ),
                1*u.dimensionless)).to(u.dimensionless)


# @u.wraps(None, [u.m, u.kg/u.m**3, u.kg/u.m**3, u.kg/u.m**3, None,
#                 None, None, u.dimensionless], False)
@ut.list_handler()
def pacl_term(DiamTube, ConcClay, ConcAl, ConcNatOrgMat, NatOrgMat,
              coag, material, RatioHeightDiameter):
    """Return the fraction of the surface area that is covered with coagulant
    that is not covered with humic acid.

    :param DiamTube: Diameter of the dosing tube
    :type Diamtube: float
    :param ConcClay: Concentration of clay in solution
    :type ConcClay: float
    :param ConcAl: Concentration of alumninum in solution
    :type ConcAl: float
    :param ConcNatOrgMat: Concentration of natural organic matter in solution
    :type ConcNatOrgMat: float
    :param NatOrgMat: type of natural organic matter, e.g. floc_model.HumicAcid
    :type NatOrgMat: floc_model.Material
    :param coag: Type of coagulant in solution, e.g. floc_model.PACl
    :type coag: floc_model.Material
    :param material: Type of clay in suspension, e.g. floc_model.Clay
    :type material: floc_model.Material
    :param RatioHeightDiameter: Dimensionless ratio of clay height to clay diameter
    :type RatioHeightDiameter: float

    :return: fraction of the surface area that is covered with coagulant that is not covered with humic acid
    :rtype: float
    """
    return (gamma_coag(ConcClay, ConcAl, coag, material, DiamTube,
                       RatioHeightDiameter)
            * (1 - gamma_humic_acid_to_coag(ConcAl, ConcNatOrgMat,
                                            NatOrgMat, coag))
            ).to(u.dimensionless)


# @u.wraps(None, [u.m, u.kg/u.m**3, u.kg/u.m**3, u.kg/u.m**3,
#                 None, None, None, u.dimensionless], False)
@ut.list_handler()
def alpha_pacl_clay(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                    NatOrgMat, coag, material, RatioHeightDiameter):
    """"""
    PAClTerm = pacl_term(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                         NatOrgMat, coag, material, RatioHeightDiameter)
    return (2 * (PAClTerm * (1 - gamma_coag(ConcClay, ConcAl, coag, material,
                                            DiamTube, RatioHeightDiameter)))).to(u.dimensionless)


# @u.wraps(None, [u.m, u.kg/u.m**3, u.kg/u.m**3, u.kg/u.m**3,
#                 None, None, None, u.dimensionless], False)
@ut.list_handler()
def alpha_pacl_pacl(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                    NatOrgMat, coag, material, RatioHeightDiameter):
    """"""
    PAClTerm = pacl_term(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                         NatOrgMat, coag, material, RatioHeightDiameter)
    return PAClTerm ** 2


# @u.wraps(None, [u.m, u.kg/u.m**3, u.kg/u.m**3, u.kg/u.m**3,
#                 None, None, None, u.dimensionless], False)
@ut.list_handler()
def alpha_pacl_nat_org_mat(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                           NatOrgMat, coag, material, RatioHeightDiameter):
    """"""
    PAClTerm = pacl_term(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                         NatOrgMat, coag, material, RatioHeightDiameter)
    return (2 * PAClTerm
            * gamma_coag(ConcClay, ConcAl, coag, material, DiamTube,
                         RatioHeightDiameter)
            * gamma_humic_acid_to_coag(ConcAl, ConcNatOrgMat, NatOrgMat, coag)).to(u.dimensionless)


# @u.wraps(None, [u.m, u.kg/u.m**3, u.kg/u.m**3, u.kg/u.m**3,
#                 None, None, None, u.dimensionless], False)
@ut.list_handler()
def alpha(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
          NatOrgMat, coag, material, RatioHeightDiameter):
    """"""
    return (alpha_pacl_nat_org_mat(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                                   NatOrgMat, coag, material,
                                   RatioHeightDiameter)
            + alpha_pacl_pacl(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                              NatOrgMat, coag, material, RatioHeightDiameter)
            + alpha_pacl_clay(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                              NatOrgMat, coag, material, RatioHeightDiameter)
            ).to(u.dimensionless)


# @u.wraps(None, [u.W/u.kg, u.degK, u.s, u.m,
#                 u.kg/u.m**3, u.kg/u.m**3, u.kg/u.m**3, None,
#                 None, None, u.dimensionless, u.dimensionless], False)
@ut.list_handler()
def pc_viscous(EnergyDis, Temp, Time, DiamTube,
               ConcClay, ConcAl, ConcNatOrgMat, NatOrgMat,
               coag, material, FittingParam, RatioHeightDiameter):
    """"""
    return ((3/2)
            * np.log10((2/3) * np.pi * FittingParam * Time
                       * np.sqrt(EnergyDis / (pc.viscosity_kinematic_water(Temp))
                                 )
                       * alpha(DiamTube, ConcClay, ConcAl, ConcNatOrgMat,
                               NatOrgMat, coag, material, RatioHeightDiameter)
                       * (np.pi/6)**(2/3)
                       * (material.Diameter / sep_dist_clay(ConcClay, material)
                          ) ** 2
                       + 1
                       ) * u.dimensionless
            ).to(u.dimensionless)


# @u.wraps(u.kg/u.m**3, [u.kg/u.m**3, u.kg/u.m**3, u.dimensionless, u.m,
#                        None, None, u.degK], False)
@ut.list_handler()
def dens_floc(ConcAl, ConcClay, DIM_FRACTAL, DiamTarget, coag, material, Temp):
    """Calculate floc density as a function of size."""
    WaterDensity = pc.density_water(Temp)
    return ((dens_floc_init(ConcAl, ConcClay, coag, material)
             - WaterDensity
             )
            * (material.Diameter / DiamTarget)**(3 - DIM_FRACTAL)
            + WaterDensity
            ).to(u.kg/u.m**3)


# @u.wraps(u.m/u.s, [u.kg/u.m**3, u.kg/u.m**3, None, None, u.dimensionless,
#                    u.m, u.degK], False)
@ut.list_handler()
def vel_term_floc(ConcAl, ConcClay, coag, material, DIM_FRACTAL,
                  DiamTarget, Temp):
    """Calculate floc terminal velocity."""
    WaterDensity = pc.density_water(Temp)
    return (((u.gravity * material.Diameter**2)
             / (18 * PHI_FLOC * pc.viscosity_kinematic_water(Temp))
             )
            * ((dens_floc_init(ConcAl, ConcClay, coag, material)
                - WaterDensity
                )
               / WaterDensity
               )
            * (DiamTarget / material.Diameter) ** (DIM_FRACTAL - 1)
            ).to(u.m/u.s)


# @u.wraps(u.m, [u.kg/u.m**3, u.kg/u.m**3, None, None,
#                u.dimensionless, u.m/u.s, u.degK], False)
@ut.list_handler()
def diam_floc_vel_term(ConcAl, ConcClay, coag, material,
                       DIM_FRACTAL, VelTerm, Temp):
    """Calculate floc diamter as a function of terminal velocity."""
    WaterDensity = pc.density_water(Temp)
    return (material.Diameter * (((18 * VelTerm * PHI_FLOC
                                   * pc.viscosity_kinematic_water(Temp)
                                   )
                                  / (u.gravity * material.Diameter**2)
                                  )
                                 * (WaterDensity
                                    / (dens_floc_init(ConcAl, ConcClay, coag,
                                                      material)
                                       - WaterDensity
                                       )
                                    )
                                 ) ** (1 / (DIM_FRACTAL - 1))
            ).to(u.m)


# @u.wraps(u.s, [u.W/u.kg, u.degK, u.kg/u.m**3, u.kg/u.m**3, None, None,
#                u.m, u.m, u.dimensionless, u.dimensionless],
#          False)
@ut.list_handler()
def time_col_laminar(EnergyDis, Temp, ConcAl, ConcClay, coag, material,
                     DiamTarget, DiamTube, DIM_FRACTAL, RatioHeightDiameter):
    """Calculate single collision time for laminar flow mediated collisions.

    Calculated as a function of floc size.
    """
    return (((1/6) * ((6/np.pi)**(1/3))
             * frac_vol_floc_initial(ConcAl, ConcClay, coag, material) ** (-2/3)
             * (pc.viscosity_kinematic_water(Temp) / EnergyDis) ** (1 / 2)
             * (DiamTarget / material.Diameter) ** (2*DIM_FRACTAL/3 - 2)
             )  # End of the numerator
            / (gamma_coag(ConcClay, ConcAl, coag, material, DiamTube,
                          RatioHeightDiameter)
               )  # End of the denominator
            ).to(u.s)


# @u.wraps(u.s, [u.W/u.kg, u.kg/u.m**3, u.kg/u.m**3, None, None,
#                u.m, u.dimensionless], False)
@ut.list_handler()
def time_col_turbulent(EnergyDis, ConcAl, ConcClay, coag, material,
                       DiamTarget, DIM_FRACTAL):
    """Calculate single collision time for turbulent flow mediated collisions.

    Calculated as a function of floc size.
    """
    return((1/6) * (6/np.pi)**(1/9) * EnergyDis**(-1/3) * DiamTarget**(2/3)
           * frac_vol_floc_initial(ConcAl, ConcClay, coag, material)**(-8/9)
           * (DiamTarget / material.Diameter)**((8*(DIM_FRACTAL-3)) / 9)
           ).to(u.s)


########### Kolmogorov and viscous length scales ###########

# @u.wraps(u.m, [u.W/u.kg, u.degK], False)
@ut.list_handler()
def eta_kolmogorov(EnergyDis, Temp):
    return (((pc.viscosity_kinematic_water(Temp) ** 3) / EnergyDis) ** (1 / 4)).to(u.m)


# @u.wraps(u.m, [u.W/u.kg, u.degK], False)
@ut.list_handler()
def lambda_vel(EnergyDis, Temp):
    return (RATIO_KOLMOGOROV * eta_kolmogorov(EnergyDis, Temp)).to(u.m)


# @u.wraps(u.m, [u.W/u.kg, u.degK, u.kg/u.m**3, u.kg/u.m**3, None, None,
#                u.dimensionless], False)
@ut.list_handler()
def diam_kolmogorov(EnergyDis, Temp, ConcAl, ConcClay, coag, material,
                    DIM_FRACTAL):
    """Return the size of the floc with separation distances equal to
    the Kolmogorov length and the inner viscous length scale.
    """
    return (material.Diameter
            * ((eta_kolmogorov(EnergyDis, Temp) / material.Diameter)
               * ((6 * frac_vol_floc_initial(ConcAl, ConcClay, coag, material))
                  / np.pi
                  )**(1/3)
               )**(3 / DIM_FRACTAL)
            ).to(u.m)


# @u.wraps(u.m, [u.W/u.kg, u.degK, u.kg/u.m**3, u.kg/u.m**3, None, None,
#                u.dimensionless], False)
@ut.list_handler()
def diam_vel(EnergyDis, Temp, ConcAl, ConcClay, coag, material, DIM_FRACTAL):
    return (material.Diameter
            * ((lambda_vel(EnergyDis, Temp) / material.Diameter)
               * ((6 * frac_vol_floc_initial(ConcAl, ConcClay, coag, material))
                  / np.pi
                  )**(1/3)
               )**(3/DIM_FRACTAL)
            ).to(u.m)


# @u.wraps(u.m, u.W/u.kg, False)
@ut.list_handler()
def diam_floc_max(epsMax):
    """
    .. deprecated:: 0.1.13
        diam_floc_max is deprecated and will be removed after Dec 1 2019. The
        underlying equation is under suspicion.

    Return floc size as a function of energy dissipation rate.

    Based on Ian Tse's work with floc size as a function of energy
    dissipation rate. This is for the average energy dissipation rate
    in a tube flocculator. It isn't clear how to convert this to the
    turbulent flow case. Maybe the flocs are mostly experiencing viscous
    shear. But that isn't clear. Some authors have made the case that
    floc breakup is due to viscous effects. If that is the case, then
    the results from the tube flocculator should be applicable to the
    turbulent case. We will have to account for the temporal and spatial
    variability in the turbulent energy dissipation rate. The factor of
    95 μm is based on the assumption that the ratio of the max to
    average energy dissipation rate for laminar flow is approximately 2.
    """
    # return (9.5 * 10**-5 * (1 / (epsMax)**(1/3))).to(u.m)
    raise FutureWarning("diam_floc_max is deprecated and will be removed after Dec"
                    " 1 2019. The underlying equation is under suspicion.")


# @u.wraps(u.W/u.kg, u.m, False)
@ut.list_handler()
def ener_dis_diam_floc(Diam):
    """
    .. deprecated:: 0.1.13
        ener_dis_diam_floc is deprecated and will be removed after Dec 1 2019.
        The underlying equation is under suspicion.

    Return max energy dissipation rate as a function of max floc diameter.
    """
    # return ((9.5 * 10**-5 / Diam) ** 3).to(u.W/u.kg)
    raise FutureWarning("ener_dis_diam_floc is deprecated and will be removed"
                        " after Dec 1 2019. The underlying equation is under"
                        " suspicion.")

##### Velocity gradient in tubing for lab scale laminar flow flocculators #####

# @u.wraps(1/u.s, [u.m**3/u.s, u.m], False)
@ut.list_handler()
def g_straight(PlantFlow, IDTube):
    return (64 * PlantFlow / (3 * np.pi * IDTube**3)).to(1/u.s)


# @u.wraps(None, [u.m**3/u.s, u.m, u.degK], False)
@ut.list_handler()
def reynolds_rapid_mix(PlantFlow, IDTube, Temp):
    return (4 * PlantFlow / (np.pi * IDTube
                             * pc.viscosity_kinematic_water(Temp))).to(u.dimensionless)


# @u.wraps(None, [u.m**3/u.s, u.m, u.m, u.degK], False)
@ut.list_handler()
def dean_number(PlantFlow, IDTube, RadiusCoil, Temp):
    """Return the Dean Number.

    The Dean Number is a dimensionless parameter that is the unfortunate
    combination of Reynolds and tube curvature. It would have been better
    to keep the Reynolds number and define a simple dimensionless geometric
    parameter.
    """
    return (reynolds_rapid_mix(PlantFlow, IDTube, Temp)
            * (IDTube / (2 * RadiusCoil))**(1/2)
            ).to(u.dimensionless)


# @u.wraps(1/u.s, [u.m**3/u.s, u.m, u.m, u.degK], False)
@ut.list_handler()
def g_coil(FlowPlant, IDTube, RadiusCoil, Temp):
    """We need a reference for this.

    Karen's thesis likely has this equation and the reference.
    """
    return (g_straight(FlowPlant, IDTube)
            * (1 + 0.033 *
                np.log10(dean_number(FlowPlant, IDTube, RadiusCoil, Temp)
                        ) ** 4
               ) ** (1/2)
            ).to(1/u.s)


# @u.wraps(u.s, [u.m, u.m, u.m**3/u.s], False)
@ut.list_handler()
def time_res_tube(IDTube, LengthTube, FlowPlant):
    """Calculate residence time in the flocculator."""
    return (LengthTube * np.pi * (IDTube**2 / 4) / FlowPlant).to(u.s)


# @u.wraps(None, [u.m**3/u.s, u.m, u.m, u.m, u.degK], False)
@ut.list_handler()
def g_time_res(FlowPlant, IDTube, RadiusCoil, LengthTube, Temp):
    """G Residence Time calculated for a coiled tube flocculator."""
    return (g_coil(FlowPlant, IDTube, RadiusCoil, Temp)
            * time_res_tube(IDTube, LengthTube, FlowPlant)
            ).to(u.dimensionless)



from aguaclara.research.procoda_parser import *
from aguaclara.core.units import u
import aguaclara.core.utility as ut
import pandas as pd
import numpy as np
from scipy import special
from scipy.optimize import curve_fit
import collections


# Carbonates
# The following code defines the carbonate system and provides functions for
# calculating Acid Neutralizing Capacity.
Kw = 10**(-14) * (u.mole/u.L)**2
K1_carbonate = 10**(-6.37)*u.mol/u.L
K2_carbonate = 10**(-10.25)*u.mol/u.L
K_Henry_CO2 = 10**(-1.5) * u.mole/(u.L*u.atm)
P_CO2 = 10**(-3.5) * u.atm


@ut.list_handler()
def invpH(pH):
    """Calculate inverse pH, i.e. hydronium ion concentration, given pH.

    :param pH: pH to be inverted
    :type pH: float

    :return: The inverse pH or hydronium ion concentration (in moles per liter)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import invpH
    >>> invpH(10)
    <Quantity(1e-10, 'mole / liter')>
    """
    return 10**(-pH)*u.mol/u.L


@ut.list_handler()
def alpha0_carbonate(pH):
    """Calculate the fraction of total carbonates in carbonic acid form (H2CO3)

    :param pH: pH of the system
    :type pH: float

    :return: Fraction of carbonates in carbonic acid form (H2CO3)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import alpha0_carbonate
    >>> round(alpha0_carbonate(10), 7)
    <Quantity(0.00015, 'dimensionless')>
    """
    alpha0_carbonate = 1/(1+(K1_carbonate/invpH(pH)) *
                            (1+(K2_carbonate/invpH(pH))))
    return alpha0_carbonate


@ut.list_handler()
def alpha1_carbonate(pH):
    """Calculate the fraction of total carbonates in bicarbonate form (HCO3-)

    :param pH: pH of the system
    :type pH: float

    :return: Fraction of carbonates in bicarbonate form (HCO3-)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import alpha1_carbonate
    >>> round(alpha1_carbonate(10), 7)
    <Quantity(0.639969, 'dimensionless')>
    """
    alpha1_carbonate = 1/((invpH(pH)/K1_carbonate) + 1 +
                          (K2_carbonate/invpH(pH)))
    return alpha1_carbonate


@ut.list_handler()
def alpha2_carbonate(pH):
    """Calculate the fraction of total carbonates in carbonate form (CO3-2)

    :param pH: pH of the system
    :type pH: float

    :return: Fraction of carbonates in carbonate form (CO3-2)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import alpha2_carbonate
    >>> round(alpha2_carbonate(10), 7)
    <Quantity(0.359881, 'dimensionless')>
    """
    alpha2_carbonate = 1/(1+(invpH(pH)/K2_carbonate) *
                            (1+(invpH(pH)/K1_carbonate)))
    return alpha2_carbonate


@ut.list_handler()
def ANC_closed(pH, total_carbonates):
    """Calculate the acid neutralizing capacity (ANC) under a closed system
    in which no carbonates are exchanged with the atmosphere during the
    experiment. Based on pH and total carbonates in the system.

    :param pH: pH of the system
    :type pH: float
    :param total_carbonates: Total carbonate concentration in the system (mole/L)
    :type total_carbonates: float

    :return: The acid neutralizing capacity of the closed system (eq/L)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import ANC_closed
    >>> from aguaclara.core.units import u
    >>> round(ANC_closed(10, 1*u.mol/u.L), 7)
    <Quantity(1.359831, 'equivalent / liter')>
    """
    return (total_carbonates * (u.eq/u.mol * alpha1_carbonate(pH) +
            2 * u.eq/u.mol * alpha2_carbonate(pH)) +
            1 * u.eq/u.mol * Kw/invpH(pH) - 1 * u.eq/u.mol * invpH(pH))


@ut.list_handler()
def ANC_open(pH):
    """Calculate the acid neutralizing capacity (ANC) calculated under an open
    system based on pH.

    :param pH: pH of the system
    :type pH: float

    :return: The acid neutralizing capacity of the closed system (eq/L)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import ANC_open
    >>> round(ANC_open(10), 7)
    <Quantity(0.0907346, 'equivalent / liter')>
    """
    return ANC_closed(pH, P_CO2*K_Henry_CO2/alpha0_carbonate(pH))


def aeration_data(DO_column, dirpath):
    """Extract the data from folder containing tab delimited
    files of aeration data. The file must be the original tab delimited file.
    All text strings below the header must be removed from these files.
    The file names must be the air flow rates with units of micromoles/s.
    An example file name would be "300.xls" where 300 is the flow rate in
    micromoles/s. The function opens a file dialog for the user to select
    the directory containing the data.

    :param DO_column: Index of the column that contains the dissolved oxygen concentration data.
    :type DO_columm: int
    :param dirpath: Path to the directory containing aeration data you want to analyze
    :type dirpath: string

    :return: collection of

        * **filepaths** (*string list*) - All file paths in the directory sorted by flow rate
        * **airflows** (*numpy.array*) - Sorted array of air flow rates with units of micromole/s
        * **DO_data** (*numpy.array list*) - Sorted list of Numpy arrays. Thus each of the numpy data arrays can have different lengths to accommodate short and long experiments
        * **time_data** (*numpy.array list*) - Sorted list of Numpy arrays containing the times with units of seconds
    """
    #return the list of files in the directory
    filenames = os.listdir(dirpath)
    #extract the flowrates from the filenames and apply units
    airflows = ((np.array([i.split('.', 1)[0] for i in filenames])).astype(np.float32))
    #sort airflows and filenames so that they are in ascending order of flow rates
    idx = np.argsort(airflows)
    airflows = (np.array(airflows)[idx])*u.umole/u.s
    filenames = np.array(filenames)[idx]

    filepaths = [os.path.join(dirpath, i) for i in filenames]
    #DO_data is a list of numpy arrays. Thus each of the numpy data arrays can have different lengths to accommodate short and long experiments
    # cycle through all of the files and extract the column of data with oxygen concentrations and the times
    DO_data=[column_of_data(i,0,DO_column,-1,'mg/L') for i in filepaths]
    time_data=[(column_of_time(i,0,-1)).to(u.s) for i in filepaths]
    aeration_collection = collections.namedtuple('aeration_results','filepaths airflows DO_data time_data')
    aeration_results = aeration_collection(filepaths, airflows, DO_data, time_data)
    return aeration_results


@ut.list_handler()
def O2_sat(P_air, temp):
    """Calculate saturaed oxygen concentration in mg/L for 278 K < T < 318 K

    :param P_air: Air pressure with appropriate units
    :type P_air: float
    :param temp: Water temperature with appropriate units
    :type temp: float

    :return: Saturated oxygen concentration in mg/L
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import O2_sat
    >>> from aguaclara.core.units import u
    >>> round(O2_sat(1*u.atm , 300*u.kelvin), 7)
    <Quantity(8.0931572, 'milligram / liter')>
    """
    fraction_O2 = 0.21
    P_O2 = P_air * fraction_O2
    return ((P_O2.to(u.atm).magnitude) *
            u.mg/u.L*np.exp(1727 / temp.to(u.K).magnitude - 2.105))


def Gran(data_file_path):
    """Extract the data from a ProCoDA Gran plot file. The file must be the original tab delimited file.

    :param data_file_path: The path to the file. If the file is in the working directory, then the file name is sufficient.

    :return: collection of

        * **V_titrant** (*float*) - Volume of titrant in mL
        * **ph_data** (*numpy.array*) - pH of the sample
        * **V_sample** (*float*) - Volume of the original sample that was titrated in mL
        * **Normality_titrant** (*float*) - Normality of the acid used to titrate the sample in mole/L
        * **V_equivalent** (*float*) - Volume of acid required to consume all of the ANC in mL
        * **ANC** (*float*) - Acid Neutralizing Capacity of the sample in mole/L
    """
    df = pd.read_csv(data_file_path, delimiter='\t', header=5)
    V_t = np.array(pd.to_numeric(df.iloc[0:, 0]))*u.mL
    pH = np.array(pd.to_numeric(df.iloc[0:, 1]))
    df = pd.read_csv(data_file_path, delimiter='\t', header=None, nrows=5)
    V_S = pd.to_numeric(df.iloc[0, 1])*u.mL
    N_t = pd.to_numeric(df.iloc[1, 1])*u.mole/u.L
    V_eq = pd.to_numeric(df.iloc[2, 1])*u.mL
    ANC_sample = pd.to_numeric(df.iloc[3, 1])*u.mole/u.L
    Gran_collection = collections.namedtuple('Gran_results', 'V_titrant ph_data V_sample Normality_titrant V_equivalent ANC')
    Gran = Gran_collection(V_titrant=V_t, ph_data=pH, V_sample=V_S,
                           Normality_titrant=N_t, V_equivalent=V_eq,
                           ANC=ANC_sample)
    return Gran


# Reactors
# The following code is for reactor responses to tracer inputs.
def CMFR(t, C_initial, C_influent):
    """Calculate the effluent concentration of a conversative (non-reacting)
    material with continuous input to a completely mixed flow reactor.

    Note: time t=0 is the time at which the material starts to flow into the
    reactor.

    :param C_initial: The concentration in the CMFR at time t=0.
    :type C_initial: float
    :param C_influent: The concentration entering the CMFR.
    :type C_influent: float
    :param t: The time(s) at which to calculate the effluent concentration. Time can be made dimensionless by dividing by the residence time of the CMFR.
    :type t: float or numpy.array

    :return: Effluent concentration
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import CMFR
    >>> from aguaclara.core.units import u
    >>> round(CMFR(0.1, 0*u.mg/u.L, 10*u.mg/u.L), 7)
    <Quantity(0.9516258, 'milligram / liter')>
    >>> round(CMFR(0.9, 5*u.mg/u.L, 10*u.mg/u.L), 7)
    <Quantity(7.9671517, 'milligram / liter')>
    """
    return C_influent * (1-np.exp(-t)) + C_initial*np.exp(-t)


def E_CMFR_N(t, N):
    """Calculate a dimensionless measure of the output tracer concentration
    from a spike input to a series of completely mixed flow reactors.

    :param t: The time(s) at which to calculate the effluent concentration. Time can be made dimensionless by dividing by the residence time of the CMFR.
    :type t: float or numpy.array
    :param N: The number of completely mixed flow reactors (CMFRS) in series. Must be greater than 1.
    :type N: int

    :return: Dimensionless measure of the output tracer concentration (concentration * volume of 1 CMFR) / (mass of tracer)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import E_CMFR_N
    >>> round(E_CMFR_N(0.5, 3), 7)
    0.7530643
    >>> round(E_CMFR_N(0.1, 1), 7)
    0.9048374
    """
    return (N**N)/special.gamma(N) * (t**(N-1))*np.exp(-N*t)

@ut.list_handler()
def E_Advective_Dispersion(t, Pe):
    """Calculate a dimensionless measure of the output tracer concentration from
    a spike input to reactor with advection and dispersion.

    :param t: The time(s) at which to calculate the effluent concentration. Time can be made dimensionless by dividing by the residence time of the CMFR.
    :type t: float or numpy.array
    :param Pe: The ratio of advection to dispersion ((mean fluid velocity)/(Dispersion*flow path length))
    :type Pe: float

    :return: dimensionless measure of the output tracer concentration (concentration * volume of reactor) / (mass of tracer)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import E_Advective_Dispersion
    >>> round(E_Advective_Dispersion(0.5, 5), 7)
    0.4774864
    """
    if t == 0:
        return 0
    else:
        return (Pe/(4*np.pi*t))**(0.5)*np.exp((-Pe*((1-t)**2))/(4*t))


def Tracer_CMFR_N(t_seconds, t_bar, C_bar, N):
    """Used by Solver_CMFR_N. All inputs and outputs are unitless. This is
    The model function, f(x, ...). It takes the independent variable as the
    first argument and the parameters to fit as separate remaining arguments.

    :param t_seconds: List of times
    :type t_seconds: float list
    :param t_bar: Average time spent in the reactor
    :type t_bar: float
    :param C_bar: Average concentration (mass of tracer)/(volume of the reactor)
    :type C_bar: float
    :param N: Number of completely mixed flow reactors (CMFRs) in series, must be greater than 1
    :type N: int

    :return: The model concentration as a function of time
    :rtype: float list

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import Tracer_CMFR_N
    >>> from aguaclara.core.units import u
    >>> Tracer_CMFR_N([1, 2, 3, 4, 5]*u.s, 5*u.s, 10*u.mg/u.L, 3)
    <Quantity([2.96358283 6.50579498 8.03352597 7.83803116 6.72125423], 'milligram / liter')>
    """
    return C_bar*E_CMFR_N(t_seconds/t_bar, N)


def Solver_CMFR_N(t_data, C_data, theta_guess, C_bar_guess):
    """Use non-linear least squares to fit the function
    Tracer_CMFR_N(t_seconds, t_bar, C_bar, N) to reactor data.

    :param t_data: Array of times with units
    :type t_data: float list
    :param C_data: Array of tracer concentration data with units
    :type C_data: float list
    :param theta_guess: Estimate of time spent in one CMFR with units.
    :type theta_guess: float
    :param C_bar_guess: Estimate of average concentration with units ((mass of tracer)/(volume of one CMFR))
    :type C_bar_guess: float

    :return: tuple of

        * **theta** (*float*)- Residence time in seconds
        * **C_bar** (*float*) - Average concentration with same units as C_bar_guess
        * **N** (*float*)- Number of CMFRS in series that best fit the data
    """
    C_unitless = C_data.magnitude
    C_units = str(C_bar_guess.units)
    t_seconds = (t_data.to(u.s)).magnitude
    # assume that a guess of 1 reactor in series is close enough to get a solution
    p0 = [theta_guess.to(u.s).magnitude, C_bar_guess.magnitude,1]
    popt, pcov = curve_fit(Tracer_CMFR_N, t_seconds, C_unitless, p0)
    Solver_theta = popt[0]*u.s
    Solver_C_bar = popt[1]*u(C_units)
    Solver_N = popt[2]
    Reactor_results = collections.namedtuple('Reactor_results','theta C_bar N')
    CMFR = Reactor_results(theta=Solver_theta, C_bar=Solver_C_bar, N=Solver_N)
    return CMFR


def Tracer_AD_Pe(t_seconds, t_bar, C_bar, Pe):
    """Used by Solver_AD_Pe. All inputs and outputs are unitless. This is the
    model function, f(x, ...). It takes the independent variable as the
    first argument and the parameters to fit as separate remaining arguments.

    :param t_seconds: List of times
    :type t_seconds: float list
    :param t_bar: Average time spent in the reactor
    :type t_bar: float
    :param C_bar: Average concentration ((mass of tracer)/(volume of the reactor))
    :type C_bar: float
    :param Pe: The Peclet number for the reactor.
    :type Pe: float

    :return: The model concentration as a function of time
    :rtype: float list

    :Examples:

    >>> from aguaclara.research.environmental_processes_analysis import Tracer_AD_Pe
    >>> from aguaclara.core.units import u
    >>> Tracer_AD_Pe([1, 2, 3, 4, 5]*u.s, 5*u.s, 10*u.mg/u.L, 5)
    <Quantity([0.25833732 3.23793989 5.8349833  6.62508831 6.30783131], 'milligram / liter')>

    """
    return C_bar*E_Advective_Dispersion(t_seconds/t_bar, Pe)


def Solver_AD_Pe(t_data, C_data, theta_guess, C_bar_guess):
    """Use non-linear least squares to fit the function
    Tracer_AD_Pe(t_seconds, t_bar, C_bar, Pe) to reactor data.

    :param t_data: Array of times with units
    :type t_data: float list
    :param C_data: Array of tracer concentration data with units
    :type C_data: float list
    :param theta_guess: Estimate of time spent in one CMFR with units.
    :type theta_guess: float
    :param C_bar_guess: Estimate of average concentration with units ((mass of tracer)/(volume of one CMFR))
    :type C_bar_guess: float

    :return: tuple of

        * **theta** (*float*)- Residence time in seconds
        * **C_bar** (*float*) - Average concentration with same units as C_bar_guess
        * **Pe** (*float*) - Peclet number that best fits the data
    """
    #remove time=0 data to eliminate divide by zero error
    t_data = t_data[1:-1]
    C_data = C_data[1:-1]
    C_unitless = C_data.magnitude
    C_units = str(C_bar_guess.units)
    t_seconds = (t_data.to(u.s)).magnitude
    # assume that a guess of 1 reactor in series is close enough to get a solution
    p0 = [theta_guess.to(u.s).magnitude, C_bar_guess.magnitude,5]
    popt, pcov = curve_fit(Tracer_AD_Pe, t_seconds, C_unitless, p0, bounds=(0.01,np.inf))
    Solver_theta = popt[0]*u.s
    Solver_C_bar = popt[1]*u(C_units)
    Solver_Pe = popt[2]
    Reactor_results = collections.namedtuple('Reactor_results', 'theta C_bar Pe')
    AD = Reactor_results(theta=Solver_theta, C_bar=Solver_C_bar, Pe=Solver_Pe)
    return AD



from aguaclara.core.units import u
import aguaclara.core.utility as ut
import numpy as np
import pandas as pd
import os

# pump rotor radius based on minimizing error between predicted and measured
# values
R_pump = 1.62 * u.cm
# empirically derived correction factor due to the fact that larger diameter
# tubing has more loss due to space smashed by rollers
k_nonlinear = 13


@ut.list_handler()
def vol_per_rev_3_stop(color="", inner_diameter=0):
    """Return the volume per revolution of an Ismatec 6 roller pump
    given the inner diameter (ID) of 3-stop tubing. The calculation is
    interpolated from the table found at
    http://www.ismatec.com/int_e/pumps/t_mini_s_ms_ca/tubing_msca2.htm.

    Note:
    1. Either input a string as the tubing color code or a number as the
    tubing inner diameter. If both are given, the function will default to using
    the color.
    2. The calculation is interpolated for inner diameters between 0.13 and 3.17
    mm. Accuracy is not guaranteed for tubes with smaller or larger diameters.

    :param color: Color code of the Ismatec 3-stop tubing
    :type color: string
    :param inner_diameter: Inner diameter of the Ismatec 3-stop tubing. Results will be most accurate for inner diameters between 0.13 and 3.17 mm.
    :type inner_diameter: float

    :return: Volume per revolution output by a 6-roller pump through the 3-stop tubing (mL/rev)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.peristaltic_pump import vol_per_rev_3_stop
    >>> from aguaclara.core.units import u
    >>> round(vol_per_rev_3_stop(color="yellow-blue"), 6)
    <Quantity(0.148846, 'milliliter / rev')>
    >>> round(vol_per_rev_3_stop(inner_diameter=.20*u.mm), 6)
    <Quantity(0.003116, 'milliliter / rev')>
    """
    if color != "":
        inner_diameter = ID_colored_tube(color)
    term1 = (R_pump * 2 * np.pi - k_nonlinear * inner_diameter) / u.rev
    term2 = np.pi * (inner_diameter ** 2) / 4
    return (term1 * term2).to(u.mL/u.rev)


@ut.list_handler()
def ID_colored_tube(color):
    """Look up the inner diameter of Ismatec 3-stop tubing given its color code.

    :param color: Color of the 3-stop tubing
    :type color: string

    :returns: Inner diameter of the 3-stop tubing (mm)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.peristaltic_pump import ID_colored_tube
    >>> from aguaclara.core.units import u
    >>> ID_colored_tube("yellow-blue")
    <Quantity(1.52, 'millimeter')>
    >>> ID_colored_tube("orange-yellow")
    <Quantity(0.51, 'millimeter')>
    >>> ID_colored_tube("purple-white")
    <Quantity(2.79, 'millimeter')>
    """
    tubing_data_path = os.path.join(os.path.dirname(__file__), "data",
        "3_stop_tubing.txt")
    df = pd.read_csv(tubing_data_path, delimiter='\t')
    idx = df["Color"] == color
    return df[idx]['Diameter (mm)'].values[0] * u.mm


@ut.list_handler()
def vol_per_rev_LS(id_number):
    """Look up the volume per revolution output by a Masterflex L/S pump
    through L/S tubing of the given ID number.

    :param id_number: Identification number of the L/S tubing. Valid numbers are 13-18, 24, 35, and 36.
    :type id_number: int

    :return: Volume per revolution output by a Masterflex L/S pump through the L/S tubing
    :rtype: float

    :Examples:

    >>> from aguaclara.research.peristaltic_pump import vol_per_rev_LS
    >>> from aguaclara.core.units import u
    >>> vol_per_rev_LS(13)
    <Quantity(0.06, 'milliliter / turn')>
    >>> vol_per_rev_LS(18)
    <Quantity(3.8, 'milliliter / turn')>
    """
    tubing_data_path = os.path.join(os.path.dirname(__file__), "data",
        "LS_tubing.txt")
    df = pd.read_csv(tubing_data_path, delimiter='\t')
    idx = df["Number"] == id_number
    return df[idx]['Flow (mL/rev)'].values[0] * u.mL/u.turn


@ut.list_handler()
def flow_rate(vol_per_rev, rpm):
    """Return the flow rate from a pump given the volume of fluid pumped per
    revolution and the desired pump speed.

    :param vol_per_rev: Volume of fluid output per revolution (dependent on pump and tubing)
    :type vol_per_rev: float
    :param rpm: Desired pump speed in revolutions per minute
    :type rpm: float

    :return: Flow rate of the pump (mL/s)
    :rtype: float

    :Examples:

    >>> from aguaclara.research.peristaltic_pump import flow_rate
    >>> from aguaclara.core.units import u
    >>> flow_rate(3*u.mL/u.rev, 5*u.rev/u.min)
    <Quantity(0.25, 'milliliter / second')>
    """
    return (vol_per_rev * rpm).to(u.mL/u.s)



"""Pipeline design functions

This module provides simple functions for creating pipeline components (pipes,
elbows, and tees), joining them together, and calculating ideal head loss/flow
rates. When they are created, pipeline components are checked against a set of
standard component sizes to ensure that they are readily constructible.

Constants:
    - ``AVAILABLE_SIZES (numpy.ndarray * u.inch)``: Set of available sizes for
      pipeline components
    - ``AVAILABLE_IDS_SCH40 (numpy.ndarray * u.inch)``: Set of available pipe
      inner diameters for SCH40 pipes
    - ``AVAILABLE_FITTING_SIZES (numpy.ndarray * u.inch)``: Set of available
      sizes for pipeline fittings
    - ``AVAILABLE_FITTING_IDS (numpy.ndarray * u.inch)``: Set of available
      inner diameters for pipeline fittings
"""
from aguaclara.core.units import unit_registry as u
from aguaclara.core import physchem as pc
from aguaclara.core import head_loss as hl
import aguaclara.core.constants as con
import aguaclara.core.materials as mats
import aguaclara.core.utility as ut
from aguaclara.design.component import Component

import pandas as pd
import numpy as np
import os.path
from abc import ABC, abstractmethod

_dir_path = os.path.dirname(__file__)
_pipe_database_path = os.path.join(_dir_path, 'data/pipe_database.csv')
with open(_pipe_database_path) as pipe_database_file:
    _pipe_database = pd.read_csv(pipe_database_file)

_fitting_database_path = \
    os.path.join(_dir_path, 'data/fitting_database.csv')
with open(_fitting_database_path) as _fitting_database_file:
    _fitting_database = pd.read_csv(_fitting_database_file)

# TODO: Once we support a Pint version that supports use with Pandas DataFrame's
# (>=0.10.0), we can assign units to DataFrame's rather than converting them to
# NumPy arrays.
_available_sizes_raw = _pipe_database.query('Used==1')['NDinch']
AVAILABLE_SIZES = np.array(_available_sizes_raw) * u.inch

_available_ids_sch40_raw = _pipe_database.query('Used==1')['ID_SCH40']
AVAILABLE_IDS_SCH40 = np.array(_available_ids_sch40_raw) * u.inch


_available_fitting_sizes_raw = _fitting_database.query('Used==1')['size']
AVAILABLE_FITTING_SIZES = np.array(_available_fitting_sizes_raw) * u.inch

_available_fitting_ids_raw = _fitting_database.query('Used==1')['id_inch']
AVAILABLE_FITTING_IDS = np.array(_available_fitting_ids_raw)* u.inch


class PipelineComponent(Component, ABC):
    """An abstract representation of pipeline components

    This abstract base class (ABC) contains common functionality for:

    #. describing and designing readily constructible pipeline components
    #. calculating the head loss of a pipeline given its flow rate, and
       vice-versa.
    #. printing the dimensions of either an individual pipe or an entire
       pipeline

    Design Inputs:
        - ``q (float * u.L/u.s)``: Flow rate (recommended, defaults to 20 L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended, defaults to
          20°C)
        - ``size (float * u.inch)``: Nominal size (recommended, defaults to 0.5
          in)
        - ``fluid_type (str)``: Fluid type. Must be 'water', 'pacl', or 'alum'
          (optional, defaults to 'water')
        - ``next (PipelineComponent)``: The next pipeline component with respect
          to the flow direction. This can be either a newly-instantiated
          ``PipelineComponent`` child class, or a variable that contains such
          a component (optional, defaults to None)
        - ``k_minor (float)``: The minor loss coefficient (k-value) (optional,
          defaults to 0)
    """
    _AVAILABLE_FLUID_TYPES = ['water', 'pacl', 'alum']

    def __init__(self, **kwargs):
        if all (key in kwargs for key in ('size', 'id')):
            raise AttributeError(
                'A PipelineComponent must be instantiated with either the size '
                'or inner diameter, but not both.'
            )

        self.size = 0.5 * u.inch
        self.fluid_type = 'water'
        self.next = None
        self.k_minor = 0

        super().__init__(**kwargs)

        self._rep_ok()

        self.size = self._get_available_size(self.size)

    @property
    def nu(self):
        """The kinematic viscosity of the fluid passing through the pipeline
        component.
        """
        if self.fluid_type == 'water':
            return pc.viscosity_kinematic_water(self.temp)
        elif self.fluid_type == 'pacl':
            print('unimplemented')
            pass
        elif self.fluid_type == 'alum':
            print('unimplemented')
            pass

    def _get_available_size(self, size):
        """Return the next larger size which is available, given the list of
        available sizes.
        """
        return ut.ceil_nearest(size, AVAILABLE_SIZES)

    @abstractmethod
    def headloss(self):
        """The head loss of this pipeline component."""
        pass

    @property
    def headloss_pipeline(self):
        """The head loss of the entire pipeline following this component."""
        if self.next is None:
            return self.headloss
        else:
            return self.headloss + self.next.headloss_pipeline

    def _set_next_components_q(self):
        """Set the flow rates of the next components in this pipeline to be
        the same as this component.
        """
        if self.next is not None:
            self.next.q = self.q
            self.next._set_next_components_q()

    def flow_pipeline(self, target_headloss):
        """Calculate the required flow through a pipeline component and all of
        its next components to reach a desired head loss.

        Args:
            - ``target_headloss (float * u.m)``: The desired head loss through
              the pipeline
        """
        if type(self) is Pipe:
            flow = pc.flow_pipe(self.id,
                    target_headloss,
                    self.l,
                    self.nu,
                    self.pipe_rough,
                    self.k_minor)
        else:
            try:
                flow = pc.flow_pipe(
                    self.next.id,
                    target_headloss,
                    self.next.l,
                    self.next.nu,
                    self.next.pipe_rough,
                    self.next.k_minor)
            except AttributeError:
                raise AttributeError('Neither of the first two components in'
                    'this pipeline are Pipe objects.')
        err = 1.0
        headloss = self.headloss_pipeline

        while abs(err) > 0.01 :
            err = (target_headloss - headloss) / (target_headloss + headloss)
            flow = flow + err * flow
            self.q = flow
            self._set_next_components_q()
            headloss = self.headloss_pipeline
        return flow.to(u.L / u.s)

    @abstractmethod
    def format_print(self):
        """The string representation of a pipeline component, disregarding other
        components in its pipeline.
        """
        pass

    def _pprint(self):
        """The pretty-printed string representation of a pipeline component
        and its next components.
        """
        if self.next is None:
            return self.format_print()
        else:
            return self.format_print() + '\n' + self.next._pprint()

    def __str__(self):
        return self._pprint()

    def __repr__(self):
        return self.__str__()

    def _rep_ok(self):
        """Ensure that the Python representation of a pipeline component is
        valid.
        """
        if self.fluid_type not in self._AVAILABLE_FLUID_TYPES:
            raise ValueError('fluid_type must be in', self._AVAILABLE_FLUID_TYPES)

        if self.next is not None:
            if type(self) is Pipe and type(self.next) not in [Elbow, Tee]:
                raise TypeError('Pipes must be connected with fittings.')
            elif type(self) in [Elbow] and type(self.next) not in [Pipe]:
                raise TypeError('Fittings must be followed by pipes.')


class Pipe(PipelineComponent):
    """Design class for a pipe

    Instantiate this class to create a readily constructible pipe and calculate
    its hydraulic features.

    ``Pipe``'s may be instantiated from a nominal size (to fit into an existing
    pipeline) or inner diameter (to follow hydraulic constraints), but not both.

    Design Inputs:
        - ``q (float * u.L/u.s)``: Flow rate (recommended, defaults to 20 L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended, defaults to
          20°C)
        - ``size (float * u.inch)``: Nominal size (recommended, defaults to 0.5
          in)
        - ``fluid_type (str)``: Fluid type. Must be 'water', 'pacl', or 'alum'
          (optional, defaults to 'water')
        - ``next (PipelineComponent)``: The next pipeline component with respect
          to the flow direction. This can be either a newly-instantiated
          ``PipelineComponent`` child class, or a variable that contains such
          a component (optional, defaults to None)
        - ``id (float * u.inch)``: Inner diameter (optional, defaults to 0.476
          in)
        - ``spec (str)``: The pipe specification. Must be one of 'sdr26',
          'sdr41', or 'sch40'. (optional, defaults to 'sdr41')
        - ``l (float * u.m)``: Length of the pipe (optional, defaults to
          1 m)
        - ``pipe_rough (float * u.mm)``: Pipe roughness (optional, defaults to
          PVC pipe roughness of 12 mm)
        - ``k_minor (float)``: The minor loss coefficient (k-value) (optional,
          defaults to 0)
    """
    AVAILABLE_SPECS = ['sdr26', 'sdr41', 'sch40']

    def __init__(self, **kwargs):
        self.id = 0.476 * u.inch
        self.spec = 'sdr41'
        self.l = 1 * u.m
        self.pipe_rough = mats.PVC_PIPE_ROUGH

        super().__init__(**kwargs)

        if 'size' in kwargs:
            self.id = self._get_id(self.size, self.spec)
        elif 'id' in kwargs:
            self.size = self._get_size(self.id, self.spec)

        self._rep_ok()

    @property
    def od(self):
        """The outer diameter of the pipe"""
        index = (
                np.abs(np.array(_pipe_database['NDinch']) - self.size.magnitude)
            ).argmin()
        return _pipe_database.iloc[index, 1] * u.inch

    def _get_size(self, id_, spec):
        """Get the size of a pipe given an inner diameter and specification.

        Args:
            - ``id_ (float * u.inch)``: Inner diameter
            - ``spec (str)``: Pipe specification
        """
        if spec[:3] == 'sdr':
            return self._get_size_sdr(id_, int(spec[3:]))
        elif spec == 'sch40':
            return self._get_size_sch40(id_)

    def _get_id(self, size, spec):
        """Get the inner diameter of a pipe given the size and specification.

        Args:
            - ``size (float * u.inch)``: Nominal size
            - ``spec (str)``: Pipe specifcation
        """
        if spec[:3] == 'sdr':
            return self._get_id_sdr(size, int(spec[3:]))
        elif spec == 'sch40':
            return self._get_id_sch40(size)

    def _get_id_sdr(self, size, sdr):
        """Get the inner diameter of a pipe given the size and SDR.

        Args:
            - ``size (float * u.inch)``: Nominal size
            - ``sdr (int)``: Standard dimension ratio
        """
        self.size = super()._get_available_size(size)
        return self.size * (sdr - 2) / sdr

    def _get_id_sch40(self, size):
        """Get the inner diameter of a SCH40 pipe.

        Args:
            - ``size (float * u.inch)``: Nominal size
        """
        self.size = super().get_available_size(size)
        myindex = (np.abs(AVAILABLE_SIZES - self.size)).argmin()
        return AVAILABLE_IDS_SCH40[myindex]

    def _get_size_sdr(self, id_, sdr):
        """Get the size of an SDR pipe.

        Args:
            - ``id_ (float * u.inch)``: Inner diameter
            - ``sdr (int)``: Standard dimension ratio
        """
        nd = super()._get_available_size((id_ * sdr) / (sdr - 2))
        self.id = self._get_id_sdr(nd, sdr)
        return nd

    def _get_size_sch40(self, id_):
        """Get the size of a SCH40 pipe.

        Args:
            - ``id_ (float * u.inch)``: Inner diameter
        """
        myindex = (np.abs(AVAILABLE_IDS_SCH40 - id_)).argmin()
        self.id = AVAILABLE_IDS_SCH40[myindex]
        return AVAILABLE_SIZES[myindex]

    def ID_SDR_all_available(self, SDR):
        """Return an array of inner diameters with a given SDR."""
        ID = []
        for i in range(len(AVAILABLE_SIZES)):
            ID.append(self._get_id_sdr(AVAILABLE_SIZES[i], SDR).magnitude)
        return ID * u.inch

    @property
    def headloss(self):
        """Return the total head loss from major and minor losses in a pipe."""
        return pc.headloss_major_pipe(
                self.q, self.id, self.l, self.nu, self.pipe_rough
            ).to(u.cm)

    def format_print(self):
        """Return the string representation of this pipe."""
        return 'Pipe: (OD: {}, Size: {}, ID: {}, Length: {}, Spec: {})'.format(
            self.od, self.size, self.id, self.l, self.spec)

    def _rep_ok(self):
        """Verify that this representation of a Pipe is valid."""
        if self.spec not in self.AVAILABLE_SPECS:
            raise AttributeError('spec must be one of:', self.AVAILABLE_SPECS)

        if self.next is not None and self.size != self.next.size:
            raise ValueError('size of the next pipeline component must be the '
            'same size as the current pipeline component')


class Elbow(PipelineComponent):
    """Design class for an Elbow

    Instantiate this class to create a readily constructible Elbow fitting and
    calculate its hydraulic features.

    ``Elbow``'s may be instantiated from a nominal size (to fit into an existing
    pipeline) or inner diameter (to follow hydraulic constraints), but not both.

    Constants:
        - ``AVAILABLE_ANGLES (int * u.deg list)``: The possible angles for this
          fitting.

    Design Inputs:
        - ``q (float * u.L/u.s)``: Flow rate (recommended, defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature
          (recommended, defaults to 20°C)
        - ``size (float * u.inch)``: The nominal size
          (recommended, defaults to 0.5 in.)
        - ``fluid_type (str)``: Fluid type. Must be 'water', 'pacl', or 'alum'
          (optional, defaults to 'water')
        - ``next (PipelineComponent)``: The next pipeline component after the
          outlet, cannot be another Elbow or a Tee fitting.
          outlet, cannot be another Elbow or a Tee fitting.
          outlet, cannot be another Elbow or a Tee fitting.
          (optional, defaults to None)
        - ``angle (float * u.deg)``: The angle of the fitting, which must be
          found in ``AVAILABLE_ANGLES`` (recommended, defaults to 90 °)
        - ``id (float * u.inch)``: The inner diameter.
          (recommended, defaults to 0.848 * u.inch)
    """
    AVAILABLE_ANGLES = [90 * u.deg, 45 * u.deg]

    def __init__(self, **kwargs):
        self.angle = 90 * u.deg
        self.id = 0.848 * u.inch

        super().__init__(**kwargs)

        self._set_k_minor()

        if 'size' in kwargs:
            self.id = self._get_id(self.size)
        elif 'id' in kwargs:
            self.size = self._get_size(self.id)

        self._rep_ok()

    def _set_k_minor(self):
        """Sets k minor"""
        if self.angle == 45 * u.deg:
            self.k_minor = hl.EL45_K_MINOR
        elif self.angle == 90 * u.deg:
            self.k_minor = hl.EL90_K_MINOR


    def _get_size(self, id_):
        """Get the size based off the inner diameter.

        Args:
            - ``id_ (float * u.inch)``: Inner diameter
        """
        myindex = (np.abs(AVAILABLE_FITTING_IDS - id_)).argmin()
        self.id = AVAILABLE_FITTING_IDS[myindex]
        return AVAILABLE_FITTING_SIZES[myindex]

    def _get_id(self, size):
        """Get the inner diameter based off the size.

        Args:
            - ``size (float * u.inch)``: Nominal Size
        """
        myindex = (np.abs(AVAILABLE_FITTING_SIZES - size)).argmin()
        self.size = AVAILABLE_FITTING_SIZES[myindex]
        return AVAILABLE_FITTING_IDS[myindex]

    @property
    def headloss(self):
        """The headloss"""
        return pc.headloss_minor_elbow(self.q, self.id, self.k_minor).to(u.cm)

    def format_print(self):
        """The string representation for an Elbow Fitting."""
        return 'Elbow: (Size: {}, ID: {}, Angle: {})'.format(
            self.size, self.id, self.angle)

    def _rep_ok(self):
        """Verify that this representation of a Elbow is valid."""
        if self.angle not in self.AVAILABLE_ANGLES:
            raise ValueError('angle must be in ', self.AVAILABLE_ANGLES)

        if self.next is not None and self.size != self.next.size:
            raise ValueError('The next component doesn\'t have the same size.')

class Tee(PipelineComponent):
    """Design class for a Tee

    Instantiate this class to create a readily constructible tee fitting and
    calculate its hydraulic features.

    ``Tee``'s may be instantiated from a nominal size (to fit into an existing
    pipeline) or inner diameter (to follow hydraulic constraints), but not both.

    Constants:
        - ``AVAILABLE_PATHS (str list)``: The available paths for the left and
          right outlet. Branch meaning the flow would turn, run meaning the flow
          stays straight, and stopper meaning there is no flow for that outlet
          due to a stopper.

    Design Inputs:
        - ``q (float * u.L / u.s)``: Flow rate (recommended, defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature
          (recommended, defaults to 20°C )
        - ``size (float * u.inch)``: The size (recommended, defaults to 0.5 in.)
        - ``fluid_type (str)``: The type of fluid flowing inside
          (optional, defaults to water)
        - ``left (PipelineComponent)``: The type of piping for the left outlet,
          cannot be an elbow or tee (recommended, defaults to None)
        - ``left_type (str)``: The type of path for the left outlet,
          can only be one of the elements in AVAILABLE_PATHS.
          can only be one of the elements in AVAILABLE_PATHS.
          can only be one of the elements in AVAILABLE_PATHS.
          (recommended, defaults to 'branch')
        - ``right (PipelineComponent)``: The type of piping for the right outlet,
          cannot be an elbow or tee. (recommended, defaults to None)
        - ``right_type (str)``: The type of path for the right outlet,
          can only be one of the elements in AVAILABLE_PATHS.
          can only be one of the elements in AVAILABLE_PATHS.
          can only be one of the elements in AVAILABLE_PATHS.
          (recommended, defaults to 'stopper')
        - ``id (float * u.inch)``: The inner diameter.
          (recommended, defaults to 0.848 * u.inch)

        """
    AVAILABLE_PATHS = ['branch', 'run', 'stopper']

    def __init__(self, **kwargs):
        self.left = None
        self.left_type = 'branch'

        self.right = None
        self.right_type = 'stopper'

        self.id = 0.848 * u.inch

        super().__init__(**kwargs)

        self._set_k_minor()
        self._set_next()

        if 'size' in kwargs:
            self.id = self._get_id(self.size)
        elif 'id' in kwargs:
            self.size = self._get_size(self.id)

        self._rep_ok()

    def _set_k_minor(self):
        """Sets k minor for the left and right outlet"""
        if self.left_type == 'branch':
            self.left_k_minor = hl.TEE_FLOW_BR_K_MINOR
        elif self.left_type == 'run':
            self.left_k_minor = hl.TEE_FLOW_RUN_K_MINOR
        elif self.left_type == 'stopper':
            self.left_k_minor = None

        if self.right_type == 'branch':
            self.right_k_minor = hl.TEE_FLOW_BR_K_MINOR
        elif self.right_type == 'run':
            self.right_k_minor = hl.TEE_FLOW_RUN_K_MINOR
        elif self.right_type == 'stopper':
            self.right_k_minor = None

    def _set_next(self):
        """Sets the next outlet as well the the type of branch for the next
        outlet.
        """
        if self.left_type == 'stopper':
            self.next = self.right
            self.next_type = self.right_type
        else:
            self.next = self.left
            self.next_type = self.left_type

    def _headloss_left(self):
        """The headloss of the left outlet"""
        return pc.headloss_minor_elbow(self.q, self.id, self.left_k_minor).to(u.cm)

    def _headloss_right(self):
        """The headloss of the right outlet"""
        return pc.headloss_minor_elbow(self.q, self.id, self.right_k_minor).to(u.cm)

    @property
    def headloss(self):
        """The headloss"""
        if self.left_type =='stopper':
            return self._headloss_right()
        else:
            return self._headloss_left()

    def _get_size(self, id_):
        """Get the nominal size based off the inner diameter

        Args:
            - ``id_ (float * u.inch)``: Inner diameter
        """
        myindex = (np.abs(AVAILABLE_FITTING_IDS - id_)).argmin()
        self.id = AVAILABLE_FITTING_IDS[myindex]
        return AVAILABLE_FITTING_SIZES[myindex]

    def _get_id(self, size):
        """Get the inner diameter based off the size.

        Args:
            - ``size (float * u.inch)``: Nominal size
        """
        myindex = (np.abs(AVAILABLE_FITTING_SIZES - size)).argmin()
        self.size = AVAILABLE_FITTING_SIZES[myindex]
        return AVAILABLE_FITTING_IDS[myindex]

    def format_print(self):
        """The string representation of this tee."""
        return 'Tee: (Size: {}, ID: {}, Next Path Type: {})'.format(
            self.size, self.id, self.next_type)

    def _rep_ok(self):
        """Verify that this representation of a Tee is valid."""
        if [self.left_type, self.right_type].count('stopper') != 1:
            raise ValueError('All tees must have one stopper.')

        if self.left_type not in self.AVAILABLE_PATHS:
            raise ValueError(
                'type of branch for left outlet must be in ',
                self.AVAILABLE_PATHS)

        if self.right_type not in self.AVAILABLE_PATHS:
            raise ValueError(
                'type of branch for right outlet must be in ',
                self.AVAILABLE_PATHS)

        if self.next is not None and self.size != self.next.size:
            raise ValueError('The next component doesn\'t have the same size.')

        if self.next is not None and type(self.next) in [Elbow, Tee]:
             raise ValueError('Tees cannot be followed by other fittings.')
